from collections.abc import Callable, Iterable
from os import environ
from os.path import exists
from pathlib import Path
from typing import Any

from cocotb_tools.runner import get_runner

from elasticai.creator.file_generation import find_project_root


def run_cocotb_sim_for_src_dir(
    src_files: Iterable[str] | Iterable[Path],
    top_module_name: str,
    cocotb_test_module: str,
    path2src: str = "",
    defines: dict | Callable[[], dict] = lambda: {},
    params: dict | Callable[[], dict] = lambda: {},
    timescale: tuple[str, str] = ("1ps", "1fs"),
    en_debug_mode: bool = True,
    waveform_save_dst: str = "",
) -> Path:
    """Function for running Verilog/VHDL Simulation using COCOTB environment
    :param src_files:           List with source files of each used Verilog/VHDL file
    :param top_module_name:     Name of the top module (from file)
    :param cocotb_test_module:  Fully qualified name of python module containing cotName of the cocotb testbench in Python
    :param path2src:            Path to the folder in which all src files are available for testing
    :param defines:             Dictionary of parameters to pass to the module [key: value, ...] - usable only in Verilog
    :param params:              Dictionary of parameters to pass to the module [key: value, ...] - value will be ignored
    :param timescale:           Tuple with Timescale value for simulation (step, accuracy)
    :param en_debug_mode:       Enable debug mode
    :param waveform_save_dst:   Path to the destination folder for saving waveform file
    :return:                    Path to folder which includes waveform file [Default: simulation output folder]
    """
    _path2src = Path(path2src)
    return run_cocotb_sim(
        src_files=[_path2src / f for f in src_files],
        top_module_name=top_module_name,
        cocotb_test_module=cocotb_test_module,
        defines=defines,
        params=params,
        timescale=timescale,
        en_debug_mode=en_debug_mode,
        waveform_save_dst=waveform_save_dst,
    )


def _build_language_map() -> dict[str, str]:
    language_mapping = {}
    for ks, language in [((".v", ".sv"), "verilog"), ((".vhd", ".vhdl"), "vhdl")]:
        for k in ks:
            language_mapping[k] = language
    return language_mapping


def run_cocotb_sim(
    src_files: Iterable[str] | Iterable[Path],
    top_module_name: str,
    cocotb_test_module: str,
    defines: dict[str, Any] | Callable[[], dict[str, Any]] = lambda: {},
    params: dict[str, Any] | Callable[[], dict[str, Any]] = lambda: {},
    timescale: tuple[str, str] = ("1ps", "1fs"),
    en_debug_mode: bool = True,
    waveform_save_dst: str = "",
    build_sim_dir: str | Path | None = None,
) -> Path:
    """Function for running Verilog/VHDL Simulation using COCOTB environment
    :param src_files:           List with source files of each used Verilog/VHDL file
    :param top_module_name:     Name of the top module (from file)
    :param cocotb_test_module:  Fully qualified name of python module containing cotName of the cocotb testbench in Python
    :param defines:             Dictionary of parameters to pass to the module [key: value, ...] - usable only in Verilog
    :param params:              Dictionary of parameters to pass to the module [key: value, ...] - value will be ignored
    :param timescale:           Tuple with Timescale value for simulation (step, accuracy)
    :param en_debug_mode:       Enable debug mode
    :param waveform_save_dst:   Path to the destination folder for saving waveform file
    :return:                    Path to folder which includes waveform file [Default: simulation output folder]
    """

    design_sources = [Path(m) for m in src_files]
    params = _normalize_dict_arg(params)
    defines = _normalize_dict_arg(defines)
    if len(design_sources) == 0:
        raise ValueError("no design sources specified")

    if any(map(lambda x: not x.exists(), design_sources)):
        raise FileNotFoundError(f"Design file does not exist {design_sources}")

    runner_mapping = {"verilog": "icarus", "vhdl": "ghdl"}
    suffix = design_sources[0].suffix
    language_mapping = _build_language_map()
    if suffix not in language_mapping:
        raise ValueError(f"File ending {suffix} not supported")

    language = language_mapping[suffix]
    runner = get_runner(runner_mapping[language])

    environ["COCOTB_RESOLVE_X"] = "ZEROS"
    environ["COCOTB_LOG_LEVEL"] = "INFO" if en_debug_mode else "WARNING"
    environ["COCOTB_REDUCED_LOG_FMT"] = "0" if en_debug_mode else "1"
    environ["MACOSX_DEPLOYMENT_TARGET"] = "15.0"

    if build_sim_dir is None:
        build_sim_dir = find_project_root() / "build_sim"
        build_sim_dir.mkdir(exist_ok=True, parents=True)
    else:
        build_sim_dir = Path(build_sim_dir)

    build_waveform_dir = (
        build_sim_dir.absolute()
        if waveform_save_dst == ""
        else Path(waveform_save_dst).absolute()
    )
    build_waveform_dir.mkdir(exist_ok=True, parents=True)
    if language == "verilog":
        # --- Building new dump file for getting vcd file
        plus_args = []
        build_args = ["-s", "cocotb_iverilog_dump_v2"]
        dump_file_src = _create_iverilog_dump_file(
            top_module_name=top_module_name,
            dump_dst=build_sim_dir,
            path=build_waveform_dir,
        )
        design_sources.append(dump_file_src)
    else:
        top_module_name = top_module_name.lower()
        build_args = []
        plus_args = [f"--vcd={build_waveform_dir / 'waveforms'}.vcd"]

    runner.build(
        sources=design_sources,
        hdl_toplevel=top_module_name,
        always=True,
        clean=False,
        waves=True,
        build_args=build_args,
        defines=defines,
        parameters=params,
        timescale=timescale,
        build_dir=build_sim_dir,
    )

    runner.test(
        hdl_toplevel=top_module_name,
        test_module=[cocotb_test_module],
        hdl_toplevel_lang=language,
        gui=False,
        plusargs=plus_args,
        waves=True,
        parameters=params,
        timescale=timescale,
        build_dir=build_sim_dir,
        test_dir=build_sim_dir,
    )
    return build_waveform_dir.absolute()


def _create_iverilog_dump_file(
    top_module_name: str, path: Path, dump_dst: Path
) -> Path:
    dumpfile_path = path / "waveforms.vcd"
    with open(dump_dst / "dump.v", "w") as f:
        f.write("module cocotb_iverilog_dump_v2();\n")
        f.write("initial begin\n")
        f.write(f'    $dumpfile("{dumpfile_path}");\n')
        f.write(f"    $dumpvars(0, {top_module_name});\n")
        f.write("end\n")
        f.write("endmodule\n")
    return dump_dst / "dump.v"


def _normalize_dict_arg(
    arg: dict[str, Any] | Callable[[], dict[str, Any]],
) -> dict[str, Any]:
    if callable(arg):
        return arg()
    return arg


def check_cocotb_test_result(result_folder_cocotb: str = "build_sim") -> bool:
    path2xml = find_project_root() / result_folder_cocotb / "results.xml"
    if not exists(path2xml):
        raise FileNotFoundError(
            "Output file of cocotb simulation 'results.xml' not found"
        )
    else:
        with open(path2xml, "r") as f:
            lines = f.readlines()

        for line in lines:
            if "failure" in line:
                return False
    return True



from enum import StrEnum

import pytest

import elasticai.creator.ir as ir
from elasticai.creator.graph import BaseGraph
from elasticai.creator.ir_transforms import SequenceReorderer
from elasticai.creator.torch2ir import (
    Implementation,
    Node,
    input_node,
    new_edge,
    output_node,
)


def reorder_blocks(impl: Implementation) -> Implementation:
    N = NodeType

    def make_nodes(*args: tuple[str, NodeType]) -> list[ir.Node]:
        return [ir.node(name=name, type=type.value) for name, type in args]

    pattern = make_nodes(
        ("conv", N.CONV1D),
        ("pool", N.MAXPOOL1D),
        ("bnorm", N.BATCHNORM1D),
        ("bin", N.BINARIZE),
        ("end", N.ANY),
    )

    replacement = make_nodes(
        ("conv", N.CONV1D),
        ("bnorm", N.BATCHNORM1D),
        ("bin", N.BINARIZE),
        ("pool", N.MAXPOOL1D),
        ("end", N.ANY),
    )

    def constraint(pattern_node: ir.Node, graph_node: ir.Node) -> bool:
        if pattern_node.type == N.ANY:
            return True
        return pattern_node.type == graph_node.type

    reorderer = SequenceReorderer(
        old_order=pattern,
        new_order=replacement,
        node_constraint=constraint,
    )

    reordered = reorderer.reorder(impl)
    return Implementation(data=reordered.data, graph=reordered.graph)


class NodeType(StrEnum):
    CONV1D = "conv1d"
    BATCHNORM1D = "batchnorm1d"
    MAXPOOL1D = "maxpool1d"
    BINARIZE = "binarize"
    ANY = "any"


class Conv1d(Node):
    in_channels: int
    out_channels: int
    kernel_size: int
    bias: bool
    groups: int
    stride: int
    padding: int

    @staticmethod
    def new(
        name: str,
        implementation: str,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        bias: bool,
        groups: int,
        stride: int,
        padding: int,
    ) -> "Conv1d":
        args = locals() | {"type": NodeType.CONV1D.value}
        return Conv1d(name, data=args)


class Batchnorm1d(Node):
    num_features: int
    affine: bool

    @staticmethod
    def new(
        name: str, implementation: str, num_features: int, affine: bool
    ) -> "Batchnorm1d":
        args = locals() | {"type": NodeType.BATCHNORM1D.value}
        return Batchnorm1d(name, data=args)


class Binarize(Node):
    @staticmethod
    def new(name: str, implementation: str) -> "Binarize":
        return Binarize(
            name=name,
            data=dict(type=NodeType.BINARIZE.value, implementation=implementation),
        )


class MaxPool1d(Node):
    kernel_size: int
    stride: int

    @staticmethod
    def new(
        name: str,
        implementation: str,
        kernel_size: int,
        stride: int,
    ) -> "MaxPool1d":
        args = locals() | {"type": NodeType.MAXPOOL1D.value}
        return MaxPool1d(name, args)


class SequentialBuilder:
    def __init__(self):
        self.clear()

    def _save_last_node(self, node: Node):
        self._last_node = node

    def clear(self) -> "SequentialBuilder":
        self.ir = Implementation(graph=BaseGraph())
        self.ir.name = "root"
        self.ir.type = "module"
        self._last_node = None
        return self

    def append(self, *nodes: Node) -> "SequentialBuilder":
        if len(nodes) > 1:
            for node in nodes:
                self.append(node)
        else:
            node = nodes[0]
            if self._last_node is None and node.type != "input":
                self.append(input_node())
            self.ir.add_node(node)
            if self._last_node is not None:
                self.ir.add_edge(new_edge(self._last_node.name, node.name))
            self._save_last_node(node)
        return self

    def build(self) -> Implementation:
        if self._last_node.type != "output":
            self.append(output_node())
        return self.ir


class TestReordering:
    # Note: the node names in the replaced subgraph are dictated by the function under test
    # We choose the same names for the nodes in the original graph, to make the test shorter

    @pytest.fixture(scope="class")
    def conv(self):
        return Conv1d.new(
            "conv",
            "conv1",
            in_channels=3,
            out_channels=64,
            kernel_size=3,
            bias=False,
            groups=1,
            stride=1,
            padding=0,
        )

    @pytest.fixture(scope="class")
    def bnorm(self):
        return Batchnorm1d.new("bnorm", "bn", num_features=64, affine=True)

    @pytest.fixture(scope="class")
    def pooling(self):
        return MaxPool1d.new("pool", "maxpool", kernel_size=2, stride=2)

    @pytest.fixture(scope="class")
    def binarize(self):
        return Binarize.new("bin", "binarize")

    @pytest.fixture(scope="class")
    def seq(
        self, conv: Conv1d, bnorm: Batchnorm1d, pooling: MaxPool1d, binarize: Binarize
    ):
        seq = SequentialBuilder()
        seq.append(conv).append(pooling).append(bnorm).append(binarize)
        return seq.build()

    @pytest.fixture(scope="class")
    def expected(
        self, conv: Conv1d, bnorm: Batchnorm1d, pooling: MaxPool1d, binarize: Binarize
    ):
        seq = SequentialBuilder()
        seq.append(conv).append(bnorm).append(binarize).append(pooling)
        return seq.build()

    @pytest.fixture(scope="class")
    def reordered(self, seq: Implementation):
        return reorder_blocks(seq)

    def test_edges_are_in_correct_order(
        self, reordered: Implementation, expected: Implementation
    ):
        assert set(reordered.edges.keys()) == set(expected.edges.keys())

    def test_conv_data_is_copied(self, reordered: Implementation, seq: Implementation):
        assert reordered.nodes["conv"] == seq.nodes["conv"]

    def test_bnorm_data_is_copied(self, reordered: Implementation, seq: Implementation):
        assert reordered.nodes["bnorm"] == seq.nodes["bnorm"]

    def test_binarize_data_is_copied(
        self, reordered: Implementation, seq: Implementation
    ):
        assert reordered.nodes["bin"] == seq.nodes["bin"]

    def test_pooling_data_is_copied(
        self, reordered: Implementation, seq: Implementation
    ):
        assert reordered.nodes["pool"] == seq.nodes["pool"]



from pathlib import Path

from elasticai.creator.ir2vhdl.testing import run_vunit_vhdl_testbenches

if __name__ == "__main__":
    run_vunit_vhdl_testbenches(
        [
            f"elasticai.creator_plugins.{name}"
            for name in (
                "padding",
                "striding_shift_register",
                "sliding_window",
                "shift_register",
                "counter",
            )
        ],
        Path(__file__).parent,
    )



from pathlib import Path

from elasticai.creator.ir2vhdl.testing import run_vunit_vhdl_testbenches

deps = ["elasticai.creator_plugins.sliding_window"]

if __name__ == "__main__":
    run_vunit_vhdl_testbenches(deps, Path(__file__).parent)



import pytest

import elasticai.creator.nn.fixed_point as nn_creator
from elasticai.creator.arithmetic import FxpArithmetic, FxpParams
from tests.integration_tests.nn.fixed_point.precomputed_routine import (
    routine_testing_precomputed_module,
)


@pytest.mark.simulation
@pytest.mark.parametrize(
    "total_bits, frac_bits, num_steps", [(6, 4, 32), (8, 4, 32), (10, 9, 64)]
)
def test_build_test_prelu_design(
    total_bits: int, frac_bits: int, num_steps: int
) -> None:
    file_name = f"TestPReLU_{total_bits}_{frac_bits}_{num_steps}"
    fxp = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    )

    dut = nn_creator.PReLU(
        total_bits=total_bits,
        frac_bits=frac_bits,
        num_steps=num_steps,
        sampling_intervall=(
            fxp.config.minimum_as_rational,
            fxp.config.maximum_as_rational,
        ),
    )
    routine_testing_precomputed_module(
        dut=dut,
        num_steps=2 * num_steps,
        fxp=fxp,
        file_name=file_name,
        file_suffix="vhd",
    )



from typing import Any, Callable

from elasticai.creator.ir2vhdl import (
    Instance,
    LogicSignal,
    LogicVectorSignal,
    NullDefinedLogicSignal,
    Signal,
    VhdlNode,
)

from .node_factory import (
    InstanceFactoryForCombinatorial,
)


@InstanceFactoryForCombinatorial.register
def clocked_combinatorial(node):
    return ClockedInstance(
        node, input_width=node.input_shape.size(), output_width=node.output_shape.size()
    )


class ClockedInstance(Instance):
    _logic_signals_with_default_suffix: tuple[str, ...] = tuple()
    _vector_signals_with_default_suffix: tuple[tuple[str, int], ...] = tuple()

    def __init__(
        self,
        node: VhdlNode,
        input_width: int,
        output_width: int,
        generic_map: dict[str, Any] | Callable[[], dict[str, Any]] = lambda: {},
    ):
        if callable(generic_map):
            generic_map = generic_map()
        port_map: dict[str, Signal] = dict(
            clk=NullDefinedLogicSignal("clk"),
            rst=NullDefinedLogicSignal("rst"),
            d_in=LogicVectorSignal(name="d_in", width=input_width),
            d_out=LogicVectorSignal(name="d_out", width=output_width),
        )
        for name, width in self._vector_signals_with_default_suffix:
            port_map[name] = LogicVectorSignal(name, width)
        for name in self._logic_signals_with_default_suffix:
            port_map[name] = LogicSignal(name)
        super().__init__(
            node,
            generic_map=node.attributes.get("generic_map", {}) | generic_map,  #  pyright: ignore
            port_map=port_map,
        )


class _ClockedCombinatorial(ClockedInstance):
    _logic_signals_with_default_suffix = ("valid_in", "valid_out")



from glob import glob
from itertools import chain
from os.path import exists
from pathlib import Path

import numpy as np
import torch

from elasticai.creator.arithmetic import FxpArithmetic
from elasticai.creator.file_generation import find_project_root
from elasticai.creator.file_generation.on_disk_path import OnDiskPath
from elasticai.creator.nn import Sequential
from elasticai.creator.testing import (
    build_report_folder_and_testdata,
    run_cocotb_sim,
)


def routine_testing_sequential_module(
    dut: Sequential,
    feat_in: int,
    fxp: FxpArithmetic,
    file_name: str,
    check_quant: bool = True,
) -> None:
    val_input = fxp.as_rational(
        torch.randint(
            low=-(2 ** (fxp.total_bits - 2)),
            high=2 ** (fxp.total_bits - 2),
            size=(20, feat_in),
        )
    )

    weights_q, bias_q = dut[0].get_params_quant()
    if check_quant:
        weights_f, bias_f = dut[0].get_params()
        error_w = (
            np.array(weights_f)
            - np.array(weights_q) * fxp.config.minimum_step_as_rational
        )
        assert np.all(np.abs(error_w) < fxp.config.minimum_step_as_rational)
        error_b = (
            np.array(bias_f) - np.array(bias_q) * fxp.config.minimum_step_as_rational
        )
        assert np.all(np.abs(error_b) < fxp.config.minimum_step_as_rational)

    dut.eval()
    with torch.no_grad():
        val_output = dut(val_input)
    output_dir = build_report_folder_and_testdata(
        dut_name=file_name,
        testdata={
            "in": fxp.cut_as_integer(val_input).int().tolist(),
            "out": fxp.round_to_integer(val_output).int().tolist(),
            "params": list(
                chain.from_iterable([a + [b] for a, b in zip(weights_q, bias_q)])
            ),
        },
    )

    output_dir_sub = Path(*output_dir.parts[len(find_project_root().parts) :])
    destination = OnDiskPath(str(output_dir_sub), parent=find_project_root())
    dut.create_design(file_name).save_to(destination)
    assert exists(output_dir / f"{file_name}.vhd")
    assert exists(output_dir / "testdata.json")

    # --- Prepare and start cocotb runner
    src_folders = [folder for folder in output_dir.iterdir() if folder.is_dir()]
    src_files = [output_dir / f"{file_name}.vhd"]
    for folder in src_folders:
        src_files.extend(glob(str(output_dir / folder / "*.vhd")))
    run_cocotb_sim(
        src_files=src_files,
        top_module_name=file_name,
        cocotb_test_module="tests.integration_tests.nn.fixed_point.sequential_tb",
        waveform_save_dst=str(output_dir),
    )



from .clocked_combinatorial import clocked_combinatorial
from .shift_register import shift_register
from .sliding_window import sliding_window
from .striding_shift_register import striding_shift_register
from .unclocked_combinatorial import unclocked_combinatorial

__all__ = [
    "clocked_combinatorial",
    "unclocked_combinatorial",
    "shift_register",
    "striding_shift_register",
    "sliding_window",
]



import logging
from collections.abc import Callable, Iterable, Iterator
from hashlib import blake2b, file_digest
from pathlib import Path


class HwFunctionIdUpdater:
    def __init__(
        self,
        build_dir: Path,
        target_file: str | Path,
        replace_id_fn: Callable[[Iterable[str], bytes], Iterator[str]],
    ):
        self._target_file = Path(target_file)
        self._build_dir = build_dir
        self._id = bytes()
        self._replace_id_fn = replace_id_fn

    def _files_from_build_dir(self) -> Iterator[Path]:
        for f in self._build_dir.glob("**/*"):
            if f.is_file():
                yield f

    def _is_not_target_file(self, file: Path) -> bool:
        return not file.samefile(self._target_file)

    def _collect_files_to_hash(self) -> Iterator[Path]:
        return filter(self._is_not_target_file, self._files_from_build_dir())

    def compute_id(self) -> None:
        logger = logging.getLogger(__name__)
        hash = _HwFunctionIdHash()

        for vhd_file in self._collect_files_to_hash():
            logger.debug(f"hashing {vhd_file.as_uri}")
            hash.update(vhd_file)
        digest = hash.digest()
        logger.debug(f"raw_digest is {digest.hex()}")

        self._id = digest

    def write_id(self) -> None:
        with open(self._target_file, "r") as f:
            code: Iterable[str] = f.readlines()
        code = self._replace_id_fn(code, self.id)
        with open(self._target_file, "w") as f:
            for line in code:
                f.write(line)
                f.write("\n")

    @property
    def id(self) -> bytes:
        return self._id


class _HwFunctionIdHash:
    SIZE: int = 16

    def __init__(self):
        self._digests = []

    def _hash(self):
        return blake2b(digest_size=self.SIZE)

    def update(self, lines: Iterable[str] | Path) -> None:
        h = self._hash()

        if isinstance(lines, Path):
            with open(lines, "rb") as f:
                self._digests.append(file_digest(f, self._hash).digest())
        else:
            for line in lines:
                h.update(line.encode())
            self._digests.append(h.digest())

    def digest(self) -> bytes:
        h = self._hash()
        for d in sorted(self._digests):
            h.update(d)
        return h.digest()



import dataclasses

import pytest

from elasticai.creator_plugins.quantized_grads.fixed_point import (
    FixedPointConfigV2,
)


def test_two_complement_fixed_point_config_is_frozen():
    config = FixedPointConfigV2(total_bits=8, frac_bits=2)
    with pytest.raises(dataclasses.FrozenInstanceError):  #
        config.total_bits = 4



from itertools import chain
from typing import Iterator

from .filter_params import FilterParameters
from .iterator_utils import batched
from .tensor import reshape_flat_CxN_groupwise


def channelwise_interleaved(length, channels):
    size_per_channel = length // channels
    indices_per_channel = batched(tuple(range(length)), size_per_channel)
    yield from chain.from_iterable(zip(*indices_per_channel))


def groupwise(length, channels, groups):
    yield from reshape_flat_CxN_groupwise(tuple(range(length)), channels, groups)


def generate_deinterleaved_indices(length, channels, channel_size):
    part_size = length // (channels * channel_size)
    by_channel = tuple(batched(tuple(range(length)), channel_size))
    by_part = tuple(batched(by_channel, part_size))
    num_parts = length // part_size // channel_size
    for c in range(part_size):
        for s in range(num_parts):
            yield from by_part[s][c]


def sliding_window(steps, size, stride):
    start = 0
    while steps > 0:
        yield range(start, start + size)
        start = start + stride
        steps = steps - 1


class GroupedFilterIndexGenerator:
    """Generates a sequence of indices as needed to apply a grouped filter kernel to a 1d signal.
    That means for a one dimensional input tensor with interleaved channels of length `spatial_size*in_channels` the generator
    can produce the correct indices for each group or step depending on the given filter parameters
    kernel size, number of input channels and stride.
    E.g.,
    >>> p = FilterParameters(in_channels=1, out_channels=1, kernel_size=2, stride=1)
    >>> g = GroupedFilterIndexGenerator(3, p)
    >>> for g_idx, group in enumerate(g.groups()):
    >>>     print("group id: ", g_idx)
    >>>     for s_idx, step in enumerate(group.steps()):
    >>>         print("step ", s_idx, ': ', list(step))
    >>>
    ... "group id: 0"
    ... "step 0: [0, 1]"
    ... "step 1: [1, 2]"
    """

    def __init__(self, params: FilterParameters):
        self._params = params

    def groups(self) -> Iterator["Group"]:
        yield from (Group(self._params, x) for x in range(self._params.groups))

    def steps(self) -> Iterator["Step"]:
        yield from (Step(self._params, x) for x in range(self._params.num_steps))

    def as_tuple_by_steps(self):
        return tuple(tuple(tuple(g) for g in step.groups()) for step in self.steps())

    def as_tuple_by_groups(self):
        return tuple(tuple(tuple(s) for s in group.steps()) for group in self.groups())


class Step:
    def __init__(self, params: FilterParameters, step_id: int):
        self._p = params
        self._step_id = step_id

    def groups(self) -> Iterator["IndicesForOperation"]:
        for group in range(self._p.groups):
            yield IndicesForOperation(
                self._p,
                self._step_id,
                group,
            )


class Group:
    def __init__(self, params: FilterParameters, group_id: int):
        self._p = params
        self._group_id = group_id

    def steps(self) -> Iterator["IndicesForOperation"]:
        for step in range(self._p.num_steps):
            yield IndicesForOperation(self._p, step, self._group_id)


class IndicesForOperation:
    def __init__(self, params: FilterParameters, step_id, group_id):
        self._p = params
        self._step = step_id
        self._group = group_id

    def __iter__(self):
        step_offset = self._p.total_skipped_inputs_per_step * self._step
        group_offset = self._group * self._p.in_channels_per_group
        offset = step_offset + group_offset
        for tap in range(self._p.kernel_size):
            for c in range(self._p.in_channels_per_group):
                yield offset + c + tap * self._p.in_channels


def unrolled_grouped_convolution(
    spatial_input_size: int,
    window_size: int,
    in_channels: int,
    groups: int,
    stride: int,
):
    """
    Generates indices for a grouped convolutional filter.
    The structure is (Steps, Groups, window_size*channels_per_group)
    """
    steps = (spatial_input_size - window_size) // stride + 1
    channels_per_group = in_channels // groups
    skipped_per_stride = stride * in_channels

    def generate_group(stride_offset, group):
        group_offset = stride_offset + group * channels_per_group

        for tap in range(window_size):
            for c in range(channels_per_group):
                yield group_offset + c + tap * in_channels

    def generate_step(step):
        for group in range(groups):
            yield generate_group(stride_offset=step * skipped_per_stride, group=group)

    for s in range(steps):
        yield generate_step(s)



from typing import Protocol, TypeVar

T = TypeVar("T")


class IrData(Protocol[T]):
    data: dict[str, dict[str, T]]


EdgeT = TypeVar("EdgeT", bound=IrData)

PNodeT = TypeVar("PNodeT", bound=IrData)
INodeT = TypeVar("INodeT", bound=IrData)
RNodeT = TypeVar("RNodeT", bound=IrData)
GNodeT = TypeVar("GNodeT", bound=IrData)

_Tcon = TypeVar("_Tcon", contravariant=True)
_TPcon = TypeVar("_TPcon", contravariant=True)


class NodeConstraintFn(Protocol[_TPcon, _Tcon]):
    def __call__(self, pattern_node: _TPcon, graph_node: _Tcon) -> bool: ...



import fnmatch

import cocotb
from cocotb.clock import Clock
from cocotb.triggers import FallingEdge, RisingEdge, Timer

from elasticai.creator.testing.cocotb_prepare import read_testdata


@cocotb.test()
async def layer_computation_test(dut):
    data = read_testdata(dut._name)
    num_feat_output = len(data["out"][-1])
    clock_period_ns = 10

    dut.enable.value = 0  # Has no impact
    dut.clock.value = 0  # has no impact
    dut.x_address.value = 0
    dut.y_address.value = 0
    dut.x.value = 0

    cocotb.start_soon(Clock(dut.clock, period=clock_period_ns, units="ns").start())
    await Timer(4 * clock_period_ns, units="ns")
    await RisingEdge(dut.clock)
    chck_test = list()
    for ite, (sig_in, ref_out) in enumerate(zip(data["in"], data["out"])):
        result = list()
        chck_ite = list()
        dut.enable.value = 1

        # --- Apply data for inference
        while dut.done.value == 0:
            dut.x.value = sig_in[dut.x_address.value]
            await FallingEdge(dut.clock)

        # --- Getting data
        await RisingEdge(dut.clock)
        for idx in range(num_feat_output):
            dut.y_address.value = idx
            for _ in range(2):
                await RisingEdge(dut.clock)
            result.append(dut.y.value.signed_integer)
            chck_ite.append(
                dut.y.value.signed_integer
                in [ref_out[idx] - 1, ref_out[idx], ref_out[idx] + 1]
            )
            for _ in range(2):
                await RisingEdge(dut.clock)
        dut.y_address.value = 0
        chck_test.extend(chck_ite)

        if not all(chck_ite):
            print(f"\n--- Run {ite} ---")
            print(f"Chck: {chck_ite}")
            print(f"Pred: {result}")
            print(f"True: {ref_out}")

        # --- Do reset
        for _ in range(2):
            await RisingEdge(dut.clock)

        dut.enable.value = 0
        for _ in range(2):
            await RisingEdge(dut.clock)

    accuracy = sum(chck_test) / len(chck_test)
    limit = 0.9
    assert accuracy >= limit, f"Accuracy of {accuracy * 100:.2f}%"


@cocotb.test()
async def layer_params_test(dut):
    layer_name = fnmatch.filter(dir(dut), "i_*linear_0")
    dut_rom = getattr(dut, layer_name[0])

    params = read_testdata(dut._name)
    dut.enable.value = 0
    dut.clock.value = 0
    dut.x_address.value = 0
    dut.y_address.value = 0
    dut.x.value = 0

    for rom_w, json_w in zip(dut_rom.rom_w.ROM.value, params["params"]):
        assert rom_w.signed_integer == json_w



from typing import cast

from elasticai.creator.file_generation.in_memory_path import InMemoryFile, InMemoryPath
from elasticai.creator.nn.fixed_point.hard_tanh.layer import HardTanh


def test_vhdl_code_matches_expected_hardtanh() -> None:
    expected = """-- This is the hard_tanh implementation for fixed point data
-- followed by the logic from pytorch:
-- https://pytorch.org/docs/stable/generated/torch.nn.Hardtanh.html
-- Version: 1.1
-- Created by: Chao
-- Changed by: AE
-- Last modified date: 2025.08.11

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity tanh is
    generic (
        DATA_WIDTH : integer := 16;
        FRAC_WIDTH : integer := 8;
        MIN_VAL : integer := -256;
        MAX_VAL : integer := 256
    );
    port (
        enable : in std_logic;
        clock  : in std_logic;
        x      : in std_logic_vector(DATA_WIDTH-1 downto 0);
        y      : out std_logic_vector(DATA_WIDTH-1 downto 0)
    );
end entity tanh;

architecture rtl of tanh is
    signal fxp_input : signed(DATA_WIDTH-1 downto 0) := (others=>'0');
    signal fxp_output : signed(DATA_WIDTH-1 downto 0) := (others=>'0');
begin
    fxp_input <= signed(x);
    y <= std_logic_vector(fxp_output);

    main_process : process(fxp_input)
    begin
        if (enable = '0') then
            fxp_output <= to_signed(0, DATA_WIDTH);
        else
            if fxp_input < to_signed(MIN_VAL, DATA_WIDTH) then
                fxp_output <= to_signed(MIN_VAL, DATA_WIDTH);
            elsif fxp_input > to_signed(MAX_VAL, DATA_WIDTH) then
                fxp_output <= to_signed(MAX_VAL, DATA_WIDTH);
            else
                fxp_output <= fxp_input;
            end if;
        end if;
    end process;
end architecture rtl;
""".splitlines()
    tanh = HardTanh(total_bits=16, frac_bits=8)
    build_path = InMemoryPath("build", parent=None)
    design = tanh.create_design("tanh")
    design.save_to(build_path)
    actual = cast(InMemoryFile, build_path["tanh"]).text
    assert actual == expected



from typing import Any, Protocol


class _NodeP(Protocol):
    implementation: str
    name: str
    type: str
    input_shape: tuple[int, int]
    output_shape: tuple[int, int]
    attributes: dict[str, Any]



from typing import Any, cast

from elasticai.creator.arithmetic import (
    FxpArithmetic,
    FxpParams,
)
from elasticai.creator.base_modules.linear import Linear as LinearBase
from elasticai.creator.nn.design_creator_module import DesignCreatorModule
from elasticai.creator.nn.fixed_point.linear.design import LinearDesign
from elasticai.creator.nn.fixed_point.math_operations import MathOperations


class Linear(DesignCreatorModule, LinearBase):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        total_bits: int,
        frac_bits: int,
        bias: bool = True,
        device: Any = None,
    ) -> None:
        self._params = FxpParams(
            total_bits=total_bits, frac_bits=frac_bits, signed=True
        )
        self._config = FxpArithmetic(self._params)
        super().__init__(
            in_features=in_features,
            out_features=out_features,
            operations=MathOperations(config=self._config),
            bias=bias,
            device=device,
        )

    def get_params(self) -> tuple[list[list[float]], list[float]]:
        bias = [0] * self.out_features if self.bias is None else self.bias.tolist()
        weights = self.weight.tolist()
        return weights, bias

    def get_params_quant(self) -> tuple[list[list[int]], list[int]]:
        weights, bias = self.get_params()
        q_weights = cast(list[list[int]], self._config.cut_as_integer(weights))
        q_bias = cast(list[int], self._config.cut_as_integer(bias))
        return q_weights, q_bias

    def create_design(self, name: str) -> LinearDesign:
        q_weights, q_bias = self.get_params_quant()
        return LinearDesign(
            frac_bits=self._config.frac_bits,
            total_bits=self._config.total_bits,
            in_feature_num=self.in_features,
            out_feature_num=self.out_features,
            weights=q_weights,
            bias=q_bias,
            name=name,
        )



import torch.nn as nn

handlers = []


def _register(fn):
    handlers.append(fn)
    return fn


@_register
def conv1d(module: nn.Conv1d) -> dict:
    return {
        "in_channels": module.in_channels,
        "out_channels": module.out_channels,
        "kernel_size": module.kernel_size,
        "stride": module.stride,
        "padding": module.padding,
        "dilation": module.dilation,
        "groups": module.groups,
        "bias": module.bias is not None,
        "padding_mode": module.padding_mode,
    }


@_register
def maxpool1d(module: nn.MaxPool1d) -> dict:
    return {
        "kernel_size": module.kernel_size,
        "stride": module.stride,
        "padding": module.padding,
        "dilation": module.dilation,
        "return_indices": module.return_indices,
        "ceil_mode": module.ceil_mode,
    }


@_register
def linear(module: nn.Linear) -> dict:
    return {
        "in_features": module.in_features,
        "out_features": module.out_features,
        "bias": module.bias is not None,
    }


@_register
def batchnorm1d(module: nn.BatchNorm1d) -> dict:
    return {
        "num_features": module.num_features,
        "affine": module.affine,
    }


@_register
def flatten(module: nn.Flatten) -> dict:
    return {}


@_register
def relu(module: nn.ReLU) -> dict:
    return {}


@_register
def sigmoid(module: nn.Sigmoid) -> dict:
    return {}



import warnings

from elasticai.creator.ir2vhdl import VhdlNode

from .clocked_combinatorial import ClockedInstance
from .node_factory import (
    InstanceFactoryForCombinatorial,
)


@InstanceFactoryForCombinatorial.register
def sliding_window(node):
    return SlidingWindowNode(node)


def _check_input_output_shape_compatibility(node):
    input_shape = node.input_shape
    output_shape = node.output_shape
    if input_shape.size() % output_shape.size() != 0:
        raise ValueError(
            "Found incompatible input output shapes for sliding_window. Total input size has to be an integer multiple of total output size, but found output={output} and input={input}.".format(
                output=output_shape, input=input_shape
            )
        )
    if output_shape.depth != input_shape.depth:
        warnings.warn(
            'Detected mismatching input output shapes for sliding_window for node "{}". Depth of output and input shape should usually be equal, but found output={} and input={}.'.format(
                node.name, output_shape, input_shape
            ),
            stacklevel=3,
        )


class SlidingWindowNode(ClockedInstance):
    _logic_signals_with_default_suffix = ("valid_in", "valid_out")

    def __init__(
        self,
        node: VhdlNode,
    ):
        _check_input_output_shape_compatibility(node)
        data_width = node.input_shape.size()
        num_points = node.output_shape.size()
        if (
            "generic_map" in node.attributes  #  pyright: ignore
            and "stride" in node.attributes["generic_map"]  #  pyright: ignore
        ):
            stride = node.attributes["generic_map"]["stride"]  # pyright: ignore
        else:
            stride = node.attributes["stride"]  # pyright: ignore

        stride = stride * node.output_shape.depth
        super().__init__(
            node,
            input_width=data_width,
            output_width=num_points,
            generic_map=dict(
                stride=stride,
                input_width=data_width,
                output_width=num_points,
            ),
        )



from itertools import chain
from typing import TypeAlias

from elasticai.creator.ir2vhdl import Code, Implementation, type_handler

from .language import Port, VHDLEntity

IOPair: TypeAlias = tuple[tuple[int, ...], tuple[int, ...]]


@type_handler
def lutron(lowered: Implementation) -> Code:
    def _iter():
        name: str = lowered.name
        d_in_width: int = lowered.attributes["input_size"]
        d_out_width: int = lowered.attributes["output_size"]
        io_pairs: tuple[IOPair, ...] = lowered.attributes["truth_table"]
        entity = VHDLEntity(
            name=name,
            port=Port(
                inputs=dict(d_in=logic_vector(d_in_width)),
                outputs=dict(d_out=logic_vector(d_out_width)),
            ),
            generics=dict(),
        )

        def rtl():
            yield from [
                f"architecture rtl of {name} is",
                "begin",
                "  process (d_in) is",
                "  begin",
                "    case d_in is",
            ]
            for _in, out in io_pairs:
                yield f'      when b"{_in}" => d_out <= b"{out}";'
            yield "      when others => d_out <= (others => 'X');"
            yield from ("    end case;", "  end process;", "end architecture;")

        yield from chain(entity.generate_entity(), ("",), rtl())

    return lowered.name, _iter()


def logic_vector(width: int | str) -> str:
    return f"std_logic_vector({width} - 1 downto 0)"



from collections.abc import Callable

import elasticai.creator.graph as gr
from elasticai.creator import ir


class PatternNode(ir.Node):
    types: set[str]


class _MultiMatcher:
    def __init__(self, pattern: ir.Implementation[PatternNode, ir.Edge]) -> None:
        self._pattern = pattern
        self.impl: ir.Implementation[ir.Node, ir.Edge] = ir.Implementation(
            graph=gr.BaseGraph(),
        )
        self.matches: list[dict[str, str]] = []

    def node_constraint(self, pattern_node: str, graph_node: str) -> bool:
        pattern_types = self._pattern.nodes[pattern_node].types
        graph_type = self.impl.nodes[graph_node].type
        if "any" in pattern_types:
            return True
        fulfilled = graph_type in pattern_types
        return fulfilled

    def update_matches(self) -> None:
        self.matches = gr.find_all_subgraphs(
            pattern=self._pattern.graph,
            graph=self.impl.graph,
            node_constraint=self.node_constraint,
        )


def build_sequential_pattern(
    *pattern: tuple[str, set[str]],
) -> ir.Implementation[PatternNode, ir.Edge]:
    p = ir.Implementation(graph=gr.BaseGraph(), node_fn=PatternNode).add_node(
        name=pattern[0][0], data=dict(types=tuple(pattern[0][1]))
    )
    for (name1, _), (name2, types2) in zip(pattern, pattern[1:]):
        p.add_node(name=name2, data=dict(types=tuple(types2)))
        p.add_edge(src=name1, dst=name2, data={})
    return p


def _rewrite_sequential(
    original: ir.Implementation,
    pattern: ir.Implementation[PatternNode, ir.Edge],
    replacement: tuple[str, ...],
) -> tuple[gr.Graph[str], list[tuple[dict[str, str], dict[str, str]]]]:
    """exchange a sequential pattern for a sequential replacement.

    The nodes named "start" and "end" are considered interface nodes and will not be replaced.

    :param original: the original implementation
    :param pattern: a sequence of tuples with node names and their possible types. Use "any" to match any type.
    :param replacement: a sequence of node names to replace the pattern
    :return: the new graph and a mapping of the new node names to the original node names
    """
    lhs = dict(start="start", end="end")
    rhs = dict(start="start", end="end")
    mm = _MultiMatcher(pattern)

    repl = original.graph.new()
    for src, dst in zip(replacement, replacement[1:]):
        repl.add_edge(src, dst)
    mm.impl = original
    mm.update_matches()
    new_graph = original.graph
    matches: list[tuple[dict[str, str], dict[str, str]]] = []
    for match in gr.get_rewriteable_matches(original.graph, mm.matches, lhs.keys()):
        new_graph, _new_names = gr.rewrite(
            original=new_graph,
            match=match,
            replacement=repl,
            rhs=rhs,
            lhs=lhs,
        )
        matches.append((match, _new_names))
    return new_graph, matches


def _build_matched_impl(
    original: ir.Implementation[ir.Node, ir.Edge],
    pattern: ir.Implementation[PatternNode, ir.Edge],
    match: dict[str, str],
) -> ir.Implementation:
    matched_impl = ir.Implementation(graph=gr.BaseGraph())
    for pattern_node, original_node in match.items():
        matched_impl.add_node(
            name=pattern_node, data=original.nodes[original_node].data
        )
        for dst in pattern.successors(pattern_node):
            matched_impl.add_edge(src=pattern_node, dst=dst, data={})
        for src in pattern.predecessors(pattern_node):
            matched_impl.add_edge(src=src, dst=pattern_node, data={})
    return matched_impl


def move_pattern_to_subimpls(
    original: ir.Implementation,
    pattern: ir.Implementation[PatternNode, ir.Edge],
    basename: str,
    replacement_data_fn: Callable[[ir.Implementation], dict[str, ir.Attribute]],
    extracted_data_fn: Callable[
        [ir.Implementation], dict[str, dict[str, ir.Attribute]]
    ] = lambda _: {},
) -> list[ir.Implementation]:
    """Move all occurences of pattern into sub implementations.

    Each pattern match will be replaced by a single node with a name derived from the basename.
    For each match, we build a new implementation with the nodes and edges of the match.

    :param original: the original implementation
    :param pattern: Each node has a set of types that will be used as match constraints. Use "any" to match any type. The special node names `"start"` and `"end"` are considered interface nodes and will not be replaced.
    :param basename: the base name for the new nodes
    :param replacement_data_fn: a function that takes the matched implementation and returns a dictionary of attributes for the newly created node.
    :param extracted_data_fn: a function that takes the matched implementation and returns a dictionary of node names and their data for the extracted implementation. Use the names from the pattern to identify nodes.
    """
    new_graph, matches = _rewrite_sequential(
        original=original,
        pattern=pattern,
        replacement=("start", basename, "end"),
    )

    extracted_subgraphs = []
    new_impl = ir.Implementation(graph=new_graph)

    for match, repl_to_new_graph_names in matches:
        extracted_impl = ir.Implementation(graph=gr.BaseGraph())
        new_node_name = repl_to_new_graph_names[basename]
        implementation_name = f"{original.name}_{new_node_name}"
        matched_impl = _build_matched_impl(original, pattern, match)
        new_impl.add_node(
            name=new_node_name,
            data=replacement_data_fn(matched_impl)[new_node_name]  # type: ignore
            | {"implementation": implementation_name},  # type: ignore
        )
        extracted_impl.name = implementation_name
        extracted_impl.type = new_impl.nodes[new_node_name].type
        for pattern_node, original_node in match.items():
            if pattern_node == "start":
                extracted_impl.add_node(name="input", data={})
            elif pattern_node == "end":
                extracted_impl.add_node(name="output", data={})
            else:
                extracted_impl.add_node(
                    name=pattern_node, data=original.nodes[original_node].data
                )
            for dst in pattern.successors(pattern_node):
                extracted_impl.add_edge(src=pattern_node, dst=dst, data={})
            for src in pattern.predecessors(pattern_node):
                extracted_impl.add_edge(src=src, dst=pattern_node, data={})
        extracted_data = extracted_data_fn(matched_impl)
        for node, data in extracted_data.items():
            extracted_impl.nodes[node].data.update(data)
        extracted_subgraphs.append(extracted_impl)

    for node in new_impl.nodes:
        if node in original.nodes:
            new_impl.add_node(original.nodes[node])
    extracted_subgraphs.append(new_impl)
    return extracted_subgraphs



import cocotb as ctb
import pytest
from cocotb.clock import Clock
from cocotb.triggers import RisingEdge
from cocotb.types import LogicArray, Range

from elasticai.creator.file_generation.resource_utils import get_file_from_package
from elasticai.creator.testing import eai_testbench

pytest_plugins = "elasticai.creator.testing.cocotb_pytest"


async def write_to_ram(dut, write_address_width, write_data_width, input_data):
    dut.write_enable.value = 1
    for address, number in enumerate(input_data):
        dut.write_address.value = LogicArray(
            address, Range(write_address_width - 1, "downto", 0)
        )
        dut.d_in.value = LogicArray(number, Range(write_data_width - 1, "downto", 0))
        await RisingEdge(dut.write_clk)
    dut.write_enable.value = 0


def start_clocks(dut):
    read_clk = Clock(dut.read_clk, 10)
    write_clk = Clock(dut.write_clk, 10)
    ctb.start_soon(read_clk.start())
    ctb.start_soon(write_clk.start())


@ctb.test()
@eai_testbench
async def can_read_from_ram(
    dut,
    write_data_width,
    write_address_width,
    write_size,
    read_data_width,
    read_address_width,
    read_size,
    input_data,
    expected,
):
    start_clocks(dut)
    await RisingEdge(dut.write_clk)
    await write_to_ram(dut, write_address_width, write_data_width, input_data)
    result = []
    dut.read_enable.value = 1

    # wait two clock cycles before starting to read
    for address in range(read_size + 2):
        if address < read_size:
            dut.read_address.value = LogicArray(
                address, Range(read_address_width - 1, "downto", 0)
            )
        await RisingEdge(dut.read_clk)
        if address > 1:
            result.append(int(dut.d_out))

    assert result == expected


@ctb.test()
@eai_testbench
async def can_write_to_ram(
    dut,
    write_data_width,
    write_address_width,
    write_size,
    read_data_width,
    read_address_width,
    read_size,
    input_data,
    expected,
):
    start_clocks(dut)
    await RisingEdge(dut.write_clk)
    await write_to_ram(dut, write_address_width, write_data_width, input_data)

    ram_content = [int(x) for x in dut.my_ram.value]
    if min(read_data_width, write_data_width) == write_data_width:
        assert ram_content == input_data
    else:
        assert ram_content == expected


@pytest.mark.simulation
@pytest.mark.parametrize(
    [
        "write_data_width",
        "write_address_width",
        "write_size",
        "read_data_width",
        "read_address_width",
        "read_size",
        "input_data",
        "expected",
    ],
    [
        (2, 2, 4, 4, 1, 2, [1, 2, 3, 0], [9, 3]),
        (4, 1, 2, 2, 2, 4, [9, 3], [1, 2, 3, 0]),
        (
            8,
            3,
            8,
            16,
            2,
            4,
            [0, 1, 1, 1, 1, 8, 15, 15],
            [2**8, 2**8 + 1, 8 * 2**8 + 1, 15 * 2**8 + 15],
        ),
    ],
)
def test_asymmetric_dual_port_bram(
    cocotb_test_fixture,
    write_data_width,
    write_address_width,
    write_size,
    read_data_width,
    read_address_width,
    read_size,
    input_data,
    expected,
):
    with get_file_from_package(
        "elasticai.creator_plugins.skeleton.vhdl", "skeleton_pkg.vhd"
    ) as f:
        cocotb_test_fixture.add_srcs(f)
        cocotb_test_fixture.run(
            params=dict(
                read_address_width=read_address_width,
                read_data_width=read_data_width,
                read_size=read_size,
                write_address_width=write_address_width,
                write_data_width=write_data_width,
                write_size=write_size,
            ),
            defines=dict(),
        )



from pathlib import Path

import pytest
import torch

from elasticai.creator.file_generation.on_disk_path import OnDiskPath
from elasticai.creator.nn.fixed_point.conv1d.layer import Conv1d
from elasticai.creator.testing.ghdl_simulation import GHDLSimulator
from elasticai.creator.testing.simulated_layer import SimulatedLayer


def create_ones_conv1d_input_list(
    batch_size: int, in_channels: int, signal_length: int
):
    return [[[1.0] * signal_length] * in_channels] * batch_size


@pytest.mark.simulation
@pytest.mark.parametrize(
    "x", ([[[0.0, 1.0, 1.0]]], [[[1.0, 1.0, 1.0]]], [[[2.0, 1.0, 0.0]]])
)
def test_verify_hw_sw_equivalence_3_inputs(
    x: list[list[list[float]]], tmp_path: Path
) -> None:
    input_data = torch.Tensor(x)
    sw_conv = Conv1d(
        total_bits=4,
        frac_bits=1,
        in_channels=1,
        out_channels=1,
        signal_length=3,
        kernel_size=2,
        bias=True,
    )
    sw_conv.weight.data = torch.ones_like(sw_conv.weight)
    sw_conv.bias.data = torch.ones_like(sw_conv.bias)

    sw_output = sw_conv(input_data)

    design = sw_conv.create_design("conv1d")
    testbench = sw_conv.create_testbench("conv1d_testbench", design)

    build_dir = OnDiskPath(name=tmp_path.name, parent=str(tmp_path.parent))
    design.save_to(build_dir.create_subpath("srcs"))
    testbench.save_to(build_dir.create_subpath("testbenches"))

    sim_layer = SimulatedLayer(testbench, GHDLSimulator, working_dir=tmp_path)
    sim_output = sim_layer(input_data)

    assert sw_output.tolist() == sim_output


@pytest.mark.simulation
@pytest.mark.parametrize(
    "x",
    (
        create_ones_conv1d_input_list(1, 2, 4),
        [[[0.5, 0.25, -1.0, 1.0], [-1.0, 1.0, -1.0, 1.0]]],
        [[[0.0, 1.0, 1.0, 0.0], [-1.0, 1.0, -1.0, 1.0]]],
    ),
)
def test_verify_hw_sw_equivalence_4_inputs(
    x: list[list[list[float]]], tmp_path: Path
) -> None:
    input_data = torch.Tensor(x)
    sw_conv = Conv1d(
        total_bits=5,
        frac_bits=2,
        in_channels=2,
        out_channels=2,
        signal_length=4,
        kernel_size=2,
        bias=True,
    )
    sw_conv.weight.data = torch.ones_like(sw_conv.weight)
    sw_conv.bias.data = torch.ones_like(sw_conv.bias)

    sw_output = sw_conv(input_data)

    design = sw_conv.create_design("conv1d")
    testbench = sw_conv.create_testbench("conv1d_testbench", design)

    build_dir = OnDiskPath(name=tmp_path.name, parent=str(tmp_path.parent))
    design.save_to(build_dir.create_subpath("srcs"))
    testbench.save_to(build_dir.create_subpath("testbenches"))

    sim_layer = SimulatedLayer(testbench, GHDLSimulator, working_dir=tmp_path)
    sim_output = sim_layer(input_data)

    assert sw_output.tolist() == sim_output



from .base_graph import BaseGraph
from .graph import Graph, ReadOnlyGraph
from .graph_iterators import (
    bfs_iter_down,
    bfs_iter_up,
    dfs_iter,
)
from .graph_rewriting import (
    DanglingEdgeError,
    RewriteResult,
    get_rewriteable_matches,
    produces_dangling_edge,
    rewrite,
)
from .name_generation import NameRegistry
from .subgraph_matching import NodeConstraintFn, find_all_subgraphs, find_subgraph

__all__ = [
    "BaseGraph",
    "Graph",
    "ReadOnlyGraph",
    "find_all_subgraphs",
    "find_subgraph",
    "bfs_iter_down",
    "bfs_iter_up",
    "dfs_iter",
    "rewrite",
    "get_rewriteable_matches",
    "NameRegistry",
    "RewriteResult",
    "produces_dangling_edge",
    "DanglingEdgeError",
    "NodeConstraintFn",
]



"""Transform PyTorch models to IR.

The translation process is highly customizable.
(`Torch2Ir`)[elasticai.creator.torch2ir.Torch2Ir] is responsible
for the translation process and will call *module handlers* to
extract attributes from modules. These attributes are then
attached to the corresponding nodes in the IR.

Each module handler is a function that takes a PyTorch module
and returns a dictionary with the extracted attributes.

The `Torch2Ir` class features some factory methods as class
methods, e.g., (`Torch2Ir.get_default_converter`)[elasticai.creator.torch2ir.Torch2Ir.get_default_converter].
Those will create a new `Torch2Ir` instance and register some
preconfigured module handlers.
However, you are free to extend or alter the behaviour of the
translation process by registering your own module handlers.
"""

from elasticai.creator.ir import edge as new_edge

from .core import Edge, Implementation, Node, input_node, new_node, output_node
from .default_handlers import handlers as default_module_handlers
from .torch2ir import Torch2Ir, get_default_converter

__all__ = [
    "Torch2Ir",
    "get_default_converter",
    "Implementation",
    "Node",
    "Edge",
    "new_node",
    "new_edge",
    "input_node",
    "output_node",
    "get_default_converter",
    "default_module_handlers",
]



from elasticai.creator.testing.ghdl_report_parsing import parse_report


def test_parse_ghdl_simulation_results_one_liner():
    simulation_output = (
        "my_test_bench.vhd:64:17:@4ps:(report note):my report message\nsimulation"
        " finished\n"
    )
    expected = [
        {
            "source": "my_test_bench.vhd",
            "line": 64,
            "column": 17,
            "time": "4ps",
            "type": "report note",
            "content": "my report message",
        }
    ]
    assert expected == parse_report(simulation_output)


def test_parse_ghdl_another_line():
    simulation_output = "A:1:2:@B:(C):D\nsimulation finished\n"
    expected = [
        {
            "source": "A",
            "line": 1,
            "column": 2,
            "time": "B",
            "type": "C",
            "content": "D",
        }
    ]
    assert expected == parse_report(simulation_output)


def test_parse_two_lines():
    simulation_output = (
        "source:1:1:@time:(type):content\nsource:1:1:@time:(type):content\nignored"
        " last line\n"
    )
    expected = [
        {
            "source": "source",
            "line": 1,
            "column": 1,
            "time": "time",
            "type": "type",
            "content": "content",
        }
    ] * 2
    assert expected == parse_report(simulation_output)


def test_put_colon_content_in_content_field():
    simulation_output = "A:1:2:@B:(C):D:e:f\n\n"
    assert "D:e:f" == parse_report(simulation_output)[0]["content"]



from typing import Any

from elasticai.creator.ir import Attribute, Edge, GraphProtocol
from elasticai.creator.ir import Implementation as _Implementation
from elasticai.creator.ir import Node as _Node


class Node(_Node):
    implementation: str


def new_node(
    name: str, type: str, implementation: str, attributes: dict[str, Any] | None = None
) -> Node:
    if attributes is None:
        attributes = {}
    return Node(name, dict(type=type, implementation=implementation) | attributes)


def input_node(attributes: dict[str, Any] | None = None) -> Node:
    return new_node("input", "input", "input", attributes)


def output_node(attributes: dict[str, Any] | None = None) -> Node:
    return new_node("output", "output", "output", attributes)


class Implementation(_Implementation[Node, Edge]):
    name: str
    type: str

    def __init__(
        self,
        *,
        graph: GraphProtocol,
        name: str | None = None,
        type: str | None = None,
        data: dict[str, Attribute] | None = None,
    ):
        super().__init__(data=data, graph=graph, node_fn=Node, edge_fn=Edge)
        if name is not None:
            self.name = name
        if type is not None:
            self.type = type



from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import (
    InProjectTemplate,
    module_to_package,
)


class Middleware:
    def save_to(self, destination: Path):
        file_names = [
            "icapInterface",
            "InterfaceStateMachine",
            "middleware",
            "spi_slave",
            "UserLogicInterface",
        ]
        for name in file_names:
            template = InProjectTemplate(
                package=module_to_package(self.__module__),
                file_name=name + ".vhd",
                parameters={},
            )
            file = destination.create_subpath(name).as_file(".vhd")
            file.write(template)



import glob
import os
import subprocess

from .ghdl_report_parsing import parse_report


class SimulationError(Exception):
    pass


class GHDLSimulator:
    """Run a simulation tool for a given `top_design` and save whatever is written to stdout
    for subsequent inspection.

    This runner uses the GHDL tool.
    The parsed content has the following keys: `("source", "line", "column", "time", "type", "content")'

    Will raise a `SimulationError` in case any of the calls to ghdl in the steps `initialize` or `run` fails.
    Args:
        workdir: typically the path to your build root, this is where we will look for vhd files
    """

    def __init__(self, workdir, top_design_name) -> None:
        self._root = workdir
        self._ghdl_dir = "ghdl_build"
        self._files = list(glob.glob("**/*.vhd", root_dir=self._root, recursive=True))
        self._standard = "08"
        self._test_bench_name = top_design_name
        self._generics: dict[str, str] = {}
        self._error_message = ""
        self._completed_process: None | subprocess.CompletedProcess = None

    def add_generic(self, **kwargs):
        self._generics.update(kwargs)

    def initialize(self):
        """Call this function once before calling `run()` and on every file change."""
        os.makedirs(f"{self._root}/{self._ghdl_dir}", exist_ok=True)
        self._load_files()
        self._compile()

    def run(self):
        """Runs the simulation and saves whatever the tool wrote to stdout.
        You're supposed to call `initialize` once, before calling `run`."""
        generic_options = [f"-g{key}={value}" for key, value in self._generics.items()]
        self._execute_command(
            self._assemble_command(["-r"])
            + ["-fsynopsys", self._test_bench_name]
            + generic_options
        )

    @property
    def _result(self) -> str:
        return self._stdout()

    def getReportedContent(self) -> list[str]:
        """Strips any information that the simulation tool added automatically to the output
        to return only the information that was printed to stdout via VHDL/Verilog statements.
        """
        parsed = parse_report(self._result)
        return list(line["content"] for line in parsed)

    def getFullReport(self) -> list[dict]:
        """Parses the output from the simulation tool, to provide a more structured representation.
        The exact content depends on the simulation tool.
        """
        return parse_report(self._result)

    def getRawResult(self) -> str:
        """Returns the raw stdout output as written by the simulation tool."""
        return self._result

    def _load_files(self):
        self._execute_command(self._assemble_command("-i") + self._files)

    def _compile(self):
        self._execute_command(
            self._assemble_command("-m") + ["-fsynopsys", self._test_bench_name]
        )

    def _execute_command(self, command):
        try:
            self._completed_process = subprocess.run(
                command, cwd=self._root, capture_output=True, check=True
            )
        except subprocess.CalledProcessError as e:
            raise SimulationError(
                f"ERROR:: executing {command}\n\tSTDERR::"
                f" {e.stderr.decode()}\n\tSTDOUT:: {e.stdout.decode()}"
            )

        self._check_for_error()

    def _check_for_error(self):
        try:
            self._completed_process.check_returncode()
        except subprocess.CalledProcessError as exception:
            error_message = self._get_error_message()
            sim_error = SimulationError(error_message)
            raise sim_error from exception

    def _get_error_message(self) -> str:
        # ghdl seems to pipe errors to stdout instead of stdin
        error_message = self._stderr()
        if error_message == "":
            error_message = self._stdout()
        return error_message

    def _stdout(self) -> str:
        if self._completed_process is not None:
            return self._completed_process.stdout.decode()
        else:
            return ""

    def _stderr(self) -> str:
        if self._completed_process is not None:
            return self._completed_process.stderr.decode()
        else:
            return ""

    def _assemble_command(self, command_flags):
        if isinstance(command_flags, str):
            command_flags = [command_flags]
        return (
            ["ghdl"] + command_flags + [f"--std={self._standard}", self._workdir_flag]
        )

    @property
    def _workdir_flag(self):
        return f"--workdir={self._ghdl_dir}"



from collections.abc import Callable
from typing import Protocol

from elasticai.creator.ir import Lowerable, LoweringPass
from elasticai.creator.plugin import PluginSymbol


class DummyLowerable(Lowerable, Protocol):
    more_data: list[str]


class CallablePluginSymbol(PluginSymbol):
    def __init__(self, fn: Callable[[DummyLowerable], str]):
        self._fn = fn

    def load_into(self, receiver: LoweringPass) -> None:
        receiver.register(self._fn.__name__)(self._fn)

    def __call__(self, x: DummyLowerable) -> str:
        return self._fn(x)


def make_plugin_symbol(fn: Callable[[DummyLowerable], str]) -> PluginSymbol:
    return CallablePluginSymbol(fn)


@make_plugin_symbol
def convolution(x: DummyLowerable) -> str:
    return "_".join(x.more_data)



from typing import cast

import torch

from elasticai.creator.arithmetic import FxpArithmetic
from elasticai.creator.base_modules.conv1d import MathOperations as Conv1dOps
from elasticai.creator.base_modules.linear import MathOperations as LinearOps
from elasticai.creator.base_modules.lstm_cell import MathOperations as LSTMOps
from elasticai.creator.nn.fixed_point.fxp_round_cut import (
    CutToFixedPoint,
    RoundToFixedPoint,
)


class MathOperations(LinearOps, Conv1dOps, LSTMOps):
    def __init__(self, config: FxpArithmetic) -> None:
        self.config = config

    def _clamp(self, a: torch.Tensor) -> torch.Tensor:
        return torch.clamp(
            input=a,
            min=self.config.minimum_as_rational,
            max=self.config.maximum_as_rational,
        )

    def _cut(self, a: torch.Tensor) -> torch.Tensor:
        return cast(torch.Tensor, CutToFixedPoint.apply(a, self.config))

    def _round(self, a: torch.Tensor) -> torch.Tensor:
        return cast(torch.Tensor, RoundToFixedPoint.apply(a, self.config))

    def round(self, a: torch.Tensor) -> torch.Tensor:
        return self._round(self._clamp(a))

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self._clamp(torch.add(a, b))

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(torch.matmul(a, b))

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(torch.mul(a, b))

    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        return self._cut(self._clamp(a))



import torch

from elasticai.creator_plugins.quantized_grads.base_modules import Linear
from elasticai.creator_plugins.quantized_grads.fixed_point import (
    FixedPointConfigV2,
    QuantizeForwHTE,
    QuantizeParamToFixedPointHTE,
    QuantizeParamToFixedPointStochastic,
)


def test_conv1d_fxp_init():
    conf = FixedPointConfigV2(8, 3)

    l = Linear(
        QuantizeForwHTE(conf),
        2,
        3,
        QuantizeParamToFixedPointHTE(conf),
        True,
        bias_quantization=QuantizeParamToFixedPointStochastic(conf),
    )

    # print(l.)
    x = torch.randn(1, 1, 2)
    l(x)



from .cocotb_prepare import build_report_folder_and_testdata, read_testdata
from .cocotb_pytest import cocotb_test_fixture, eai_testbench
from .cocotb_runner import (
    check_cocotb_test_result,
    run_cocotb_sim,
    run_cocotb_sim_for_src_dir,
)
from .ghdl_report_parsing import parse_report
from .ghdl_simulation import GHDLSimulator
from .simulated_layer import SimulatedLayer, Testbench

__all__ = [
    "run_cocotb_sim",
    "run_cocotb_sim_for_src_dir",
    "check_cocotb_test_result",
    "build_report_folder_and_testdata",
    "read_testdata",
    "GHDLSimulator",
    "parse_report",
    "Testbench",
    "SimulatedLayer",
    "make_cocotb_pytest",
    "cocotb_test_fixture",
    "eai_testbench",
]



import re
from collections.abc import Callable
from typing import cast

import pytest

from elasticai.creator.file_generation.in_memory_path import InMemoryFile, InMemoryPath
from elasticai.creator.vhdl.shared_designs.rom.design import Rom


def extract_rom_values(text: str | list[str]) -> tuple[str, ...]:
    if not isinstance(text, list):
        text = [text]
    values: tuple[str, ...] = tuple()
    for line in text:
        match = re.match(
            r'.*\("([a-f0-9]+(",\s?"[a-f0-9]+)*)"\)',
            line,
        )
        if match is not None:
            array = match.group(1)
            values = tuple(re.split(r'(?:",\s?")', array))

    return values


class TestExtractRomValues:
    @pytest.mark.parametrize(
        ("values_to_extract", "values"),
        [
            ('"0"', ("0",)),
            ('"0", "1"', ("0", "1")),
            ('"10101111", "10111011"', ("10101111", "10111011")),
            ('"10101111","10111011"', ("10101111", "10111011")),
        ],
    )
    def test_extracted_rom_values(
        self, values_to_extract: str, values: tuple[str]
    ) -> None:
        test_line = f"signal ROM : some_name_array_t:=({values_to_extract})"
        assert extract_rom_values(test_line) == values

    def test_can_handle_iteration_over_multiple_lines(self):
        test_lines = [
            "some text",
            'some text that is comparable to array ("0", "1")',
            'some_other_name_array_t:=("10101111", "10111011")',
        ]
        assert extract_rom_values(test_lines) == ("10101111", "10111011")


@pytest.fixture
def rom_name() -> str:
    return "test_rom"


@pytest.fixture
def build_root() -> InMemoryPath:
    return InMemoryPath("build", parent=None)


@pytest.fixture
def rom_code(
    rom_name: str, build_root: InMemoryPath
) -> Callable[[list[int]], list[str]]:
    destination = build_root

    def generate_code(values_as_integers: list[int]) -> list[str]:
        rom = Rom(name=rom_name, data_width=8, values_as_integers=values_as_integers)
        rom.save_to(destination)
        return cast(InMemoryFile, build_root[rom_name]).text

    return generate_code


@pytest.fixture
def address_width(rom_code) -> Callable[[list[int]], int]:
    def extract(values_as_integers: list[int]) -> int:
        text = rom_code(values_as_integers)
        actual = 0
        for line in text:
            match = re.match(r"\s+ROM_ADDR_WIDTH : integer :=\s+(\d+);", line)
            if match:
                actual = int(match.group(1))
                break
        return actual

    return extract


class TestRomDesign:
    @pytest.mark.parametrize(
        ("values_as_integers", "expected_rom_values"),
        [
            ([0, 0], ("00000000", "00000000")),
            ([-81, -69], ("10101111", "10111011")),
        ],
    )
    def test_generating_correct_rom_values(
        self,
        rom_code: Callable[[list[int]], list[str]],
        values_as_integers: list[int],
        expected_rom_values: tuple[str, ...],
    ) -> None:
        code = rom_code(values_as_integers)
        assert extract_rom_values(code) == expected_rom_values

    @pytest.mark.parametrize(
        ("values_as_integers", "expected_rom_values"),
        [
            ([1] * 3, ("00000001", "00000001", "00000001", "00000000")),
            ([1] * 18, tuple(["00000001"] * 18 + ["00000000"] * 14)),
        ],
    )
    def test_rom_values_are_padded_correctly(
        self,
        rom_code: Callable[[list[int]], list[str]],
        values_as_integers: list[int],
        expected_rom_values: list[str],
    ) -> None:
        code = rom_code(values_as_integers)
        assert extract_rom_values(code) == expected_rom_values

    @pytest.mark.parametrize(
        ("values_as_integers", "expected_address_width"),
        [
            ([0, 0], 1),
            ([0, 0, 0, 0, 0], 3),
        ],
    )
    def test_address_width_is_set_correctly(
        self,
        address_width: Callable[[list[int]], int],
        values_as_integers: list[int],
        expected_address_width: int,
    ) -> None:
        assert address_width(values_as_integers) == expected_address_width



from typing import Any, Protocol

import torch

from elasticai.creator.base_modules.math_operations import Add, MatMul, Quantize


class MathOperations(Quantize, Add, MatMul, Protocol): ...


class Linear(torch.nn.Linear):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        operations: MathOperations,
        bias: bool,
        device: Any = None,
        dtype: Any = None,
    ) -> None:
        super().__init__(
            in_features=in_features,
            out_features=out_features,
            bias=bias,
            device=device,
            dtype=dtype,
        )
        self._operations = operations

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        weight = self._operations.quantize(self.weight)
        if self.bias is None:
            return self._operations.matmul(x, weight.T)
        else:
            return self._operations.add(
                self._operations.matmul(x, weight.T),
                self._operations.quantize(self.bias),
            )



from collections.abc import MutableMapping
from types import GenericAlias, MappingProxyType, resolve_bases
from typing import Any, TypeAlias, cast

from .required_field import SimpleRequiredField, is_required_field


class IrDataMeta(type):
    """
    This implementation tries to find a good compromise between
    avoiding boiler plate code and compatibility with
    static type checkers.

    The metaclass provides automatic generation of an init function
    and will convert any type annotations with `Attribute` types into corresponding
    `MandatoryField`s.
    To **not** generate an init function define your class like this

    ```python
    class C(metaclass=IrDataMeta, create_init=False):
      name: str
    ```

    We recommend inheriting from the `IrData` class instead as this provides
    useful checks for data integrity and an init function that will be detected
    by static type checkers.

    IMPORTANT: Type annotations for names starting with `_` are excluded from this!


    NOTE:
    Other options of implementation have been considered and tried:

    - dynamically add fields and init function after class creation
      - (-)(-) neither fields nor data attribute detected by static type checkers
      - (-) __set_name__ callbacks are not triggered
      - (+) easier to understand
      - (-) __slots__ need to be defined manually by the user, which will certainly lead
        to hard to find bugs.
    - a decorator that builds a new class taking the decorated class as a prototype
      - (+)(-) type hints for fields are picked up
      - (-) a lot of boiler plate to make type checkers detect init and other attributes
      - (+) if it wasn't for the type checks, this would be easy to understand and maintain
        and have the least amount of boiler plate
    """

    @property
    def required_fields(self) -> MappingProxyType[str, type]:
        """
        In `IrData` and its children this will be available
        as a class property. We add it here because python3.13 removed
        the ability to combine classmethod and property
        """
        return self._fields.keys().mapping  # type: ignore

    @classmethod
    def __prepare__(cls, name, bases, create_init=False, **kwds):
        """
        called before parsing class body and thus before metaclass instantiation,
        hence this is a classmethod
        """
        namespace = super().__prepare__(name, bases, **kwds)
        cls.__add_data_slot(namespace)
        if create_init:
            cls.__add_init(namespace)
        namespace["_fields"] = {}
        return namespace

    @classmethod
    def __add_init(cls, namespace):
        def init(self, data):
            self.data = data

        namespace["__init__"] = init

    def __new__(
        cls, name: str, bases: tuple[type, ...], namespace: dict[str, Any], **kwds
    ) -> "IrDataMeta":
        """
        Create and return a new IrDataMeta object.
        This is the type for new IrData classes.
        Fields defined in the namespace are fields of the class, not fields
        of the instantiated object. E.g., the deriving class will feature
        an `__init__` method, the owner of this method is the class, i.e., an object
        of type `IRDataMeta`. When accessing such a method attribute via the `dot`
        from an object with metaclass `IrDataMeta`, the method will be bound
        to that specific instance on the fly.

        As such, when dynamically creating the mandatory fields below,
        this happens once per class definition, not per instantiation
        of such a class.

        """
        inherited_fields = cls.__get_inherited_fields(bases)
        generated_fields = cls.__get_fields_to_be_generated(namespace)
        user_defined_fields = cls.__get_user_defined_fields_from_annotations(namespace)
        for name in generated_fields:
            namespace[name] = SimpleRequiredField()
        namespace["_fields"] = inherited_fields | generated_fields | user_defined_fields
        return super().__new__(cls, name, bases, namespace)

    @staticmethod
    def __get_inherited_fields(bases):
        inherited_fields = {}
        for o in resolve_bases(bases):
            if isinstance(o, IrDataMeta):
                inherited_fields.update(o._fields)  # type: ignore
        return inherited_fields

    @staticmethod
    def __add_data_slot(namespace: MutableMapping[str, object]) -> None:
        namespace["__slots__"] = ("data",) + cast(
            tuple, namespace.get("__slots__", tuple())
        )

    @staticmethod
    def __get_user_defined_fields_from_annotations(
        namespace: dict[str, Any],
    ) -> dict[str, type]:
        annotations = namespace.get("__annotations__", {})
        fields = dict()
        for name, annotation in annotations.items():
            if not name.startswith("_") and name in namespace:
                item = namespace[name]
                if is_required_field(item):
                    stored_type = annotation.__args__[0]
                    fields[name] = stored_type
        return cast(dict[str, type], fields)

    @staticmethod
    def __get_fields_to_be_generated(
        namespace: dict[str, Any],
    ) -> dict[str, type]:
        annotations = namespace.get("__annotations__", {})
        fields: dict[str, type] = dict()
        for name, annotation in annotations.items():
            if name.startswith("_") or name in namespace:
                continue
            fields[name] = IrDataMeta.__get_annotations_type(annotation)
        return fields

    @staticmethod
    def __get_annotations_type(annotation) -> type | TypeAlias:
        if isinstance(annotation, str):
            annotation = eval(annotation, globals(), locals())
        if isinstance(annotation, GenericAlias):
            _type = annotation.__origin__
        else:
            _type = annotation
        return _type



import dataclasses
from abc import abstractmethod
from typing import Any, Iterator, Protocol


class Definable(Protocol):
    @property
    @abstractmethod
    def name(self) -> str: ...

    @abstractmethod
    def define(self) -> Iterator[str]: ...


class Node(Protocol):
    implementation: str
    name: str
    type: str
    input_shape: tuple[int, int]
    output_shape: tuple[int, int]
    attributes: dict[str, Any]


class VHDLNodeWrapper:
    def __init__(
        self,
        node: Node,
        generic_map: dict[str, str],
        port_map: dict[str, Definable],
    ):
        self._node = node
        self._generics: dict[str, str] = {k.lower(): v for k, v in generic_map.items()}
        self.port_map = port_map

    def add_signal_with_suffix(
        self, signal: Definable, prefix: str | None = None
    ) -> None:
        suffix = self._node.name
        prefix = signal.name if prefix is None else prefix
        self.port_map.update(
            {signal.name: dataclasses.replace(signal, name=f"{prefix}_{suffix}")}
        )

    @property
    def name(self) -> str:
        return self._node.name

    @property
    def implementation(self) -> str:
        return self._node.implementation

    def define_signals(self) -> Iterator[str]:
        for s in self.port_map.values():
            yield from s.define()

    def generate_entity(self) -> Iterator[str]:
        yield ""

    def instantiate(self) -> Iterator[str]:
        yield from (f"{self.name}: entity work.{self.implementation}(rtl) ",)
        generics = tuple(self._generics.items())
        if len(generics) > 0:
            yield "generic map ("
            for key, value in generics[:-1]:
                yield f"  {key.upper()} => {value},"
            for g in generics[-1:]:
                yield f"  {g[0].upper()} => {g[1]}"
            yield "  )"
        port_map = tuple(self.port_map.items())
        yield "  port map ("

        for k, v in port_map[:-1]:
            yield f"    {k} => {v},"
        for k, v in port_map[-1:]:
            yield f"    {k} => {v}"

        yield "  );"



import torch
from torch import Tensor


class Round(torch.autograd.Function):
    """
    Round deterministically to nearest neighbour with STE.
    """

    @staticmethod
    def forward(ctx, x, *args, **kwargs):
        return torch.round(x)

    @staticmethod
    def backward(ctx, *grad_output):
        return grad_output


def round_tensor(x: Tensor) -> Tensor:
    return Round.apply(x)


def _clamp(
    number: Tensor, minimum_as_rational: Tensor, maximum_as_rational: Tensor
) -> Tensor:
    """
    Clamp values to twos complement.
    """
    return torch.clamp(
        number,
        minimum_as_rational,
        maximum_as_rational,
    )


def _round_to_fixed_point_hte(number: Tensor, resolution: Tensor) -> Tensor:
    """
    Implements round half to even with torch rounding function.
    """
    return round_tensor(number * resolution) / resolution


def quantize_to_fxp_hte(
    number: Tensor,
    resolution_per_int: Tensor,
    minimum_as_rational: Tensor,
    maximum_as_rational: Tensor,
) -> Tensor:
    """
    Round fixed point half to even.
    The tensor is clamped and rounded to a fixed point number.
    """
    return _clamp(
        _round_to_fixed_point_hte(number, resolution_per_int),
        minimum_as_rational,
        maximum_as_rational,
    )


def _noise(x: Tensor, resolution_per_int: Tensor) -> Tensor:
    return (torch.rand_like(x) - 0.5) / resolution_per_int


def quantize_to_fxp_stochastic(
    number: Tensor,
    resolution_per_int: Tensor,
    minimum_as_rational: Tensor,
    maximum_as_rational: Tensor,
) -> Tensor:
    return quantize_to_fxp_hte(
        number + _noise(number, resolution_per_int),
        resolution_per_int,
        minimum_as_rational,
        maximum_as_rational,
    )



from typing import Any, cast

import torch

from elasticai.creator.arithmetic import (
    FxpArithmetic,
    FxpParams,
)
from elasticai.creator.base_modules.linear import Linear as LinearBase
from elasticai.creator.nn.design_creator_module import DesignCreatorModule
from elasticai.creator.nn.fixed_point.linear.design import LinearDesign
from elasticai.creator.nn.fixed_point.math_operations import MathOperations


class BatchNormedLinear(DesignCreatorModule, torch.nn.Module):
    def __init__(
        self,
        total_bits: int,
        frac_bits: int,
        in_features: int,
        out_features: int,
        bias: bool = True,
        bn_eps: float = 1e-5,
        bn_momentum: float = 0.1,
        bn_affine: bool = True,
        device: Any = None,
    ) -> None:
        super().__init__()
        self._params = FxpParams(
            total_bits=total_bits, frac_bits=frac_bits, signed=True
        )
        self._config = FxpArithmetic(self._params)
        self._operations = MathOperations(config=self._config)
        self._linear = LinearBase(
            in_features=in_features,
            out_features=out_features,
            operations=self._operations,
            bias=bias,
            device=device,
        )
        self._batch_norm = torch.nn.BatchNorm1d(
            num_features=out_features,
            eps=bn_eps,
            momentum=bn_momentum,
            affine=bn_affine,
            track_running_stats=True,
            device=device,
        )

    @property
    def lin_weight(self) -> torch.Tensor:
        return self._linear.weight

    @property
    def lin_bias(self) -> torch.Tensor:
        return self._linear.bias

    @property
    def bn_weight(self) -> torch.Tensor:
        return self._batch_norm.weight

    @property
    def bn_bias(self) -> torch.Tensor:
        return self._batch_norm.bias

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        has_batches = x.dim() == 2
        input_shape = x.shape if has_batches else (1, -1)
        output_shape = (x.shape[0], -1) if has_batches else (-1,)

        x = x.view(*input_shape)
        x = self._linear(x)
        x = self._batch_norm(x)
        x = self._operations.quantize(x)
        return x.view(*output_shape)

    def get_params(self) -> tuple[list[list[float]], list[float]]:
        bn_mean = cast(torch.Tensor, self._batch_norm.running_mean)
        bn_variance = cast(torch.Tensor, self._batch_norm.running_var)
        bn_epsilon = self._batch_norm.eps
        lin_weight = self._linear.weight
        lin_bias = (
            torch.tensor([0] * self._linear.out_features)
            if self._linear.bias is None
            else self._linear.bias
        )

        std = torch.sqrt(bn_variance + bn_epsilon)
        weights = lin_weight / std[:, None]
        bias = (lin_bias - bn_mean) / std
        if self._batch_norm.affine:
            weights = (self._batch_norm.weight * weights.t()).t()
            bias = self._batch_norm.weight * bias + self._batch_norm.bias
        return weights.tolist(), bias.tolist()

    def get_params_quant(self) -> tuple[list[list[int]], list[int]]:
        weights, bias = self.get_params()
        q_weights = cast(list[list[int]], self._config.cut_as_integer(weights))
        q_bias = cast(list[int], self._config.cut_as_integer(bias))
        return q_weights, q_bias

    def create_design(self, name: str) -> LinearDesign:
        weights, bias = self.get_params_quant()
        return LinearDesign(
            in_feature_num=self._linear.in_features,
            out_feature_num=self._linear.out_features,
            total_bits=self._operations.config.total_bits,
            frac_bits=self._operations.config.frac_bits,
            weights=weights,
            bias=bias,
            name=name,
        )



from collections.abc import Callable

from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.vhdl.auto_wire_protocols.port_definitions import create_port
from elasticai.creator.vhdl.design.design import Design
from elasticai.creator.vhdl.design.ports import Port


class PrecomputedScalarFunction(Design):
    _template_package = module_to_package(__name__)

    def __init__(
        self,
        name: str,
        input_width: int,
        output_width: int,
        function: Callable[[int], int],
        inputs: list[int],
    ) -> None:
        super().__init__(name)
        self._input_width = input_width
        self._output_width = output_width
        self._function = function
        self._inputs = inputs
        self._template = InProjectTemplate(
            file_name="precomputed_scalar_function.tpl.vhd",
            package=self._template_package,
            parameters=dict(
                name=self.name,
                input_data_width=str(self._input_width),
                output_data_width=str(self._output_width),
            ),
        )

    def _compute_io_pairs(self) -> list[tuple[int, int]]:
        ascending_unique_inputs = sorted(set(self._inputs))
        io_pairs = []
        for input_value in ascending_unique_inputs:
            _assert_value_is_representable_with_n_bits(input_value, self._input_width)
            output_value = self._function(input_value)
            _assert_value_is_representable_with_n_bits(output_value, self._output_width)
            io_pairs.append((input_value, output_value))
        return io_pairs

    @property
    def port(self) -> Port:
        return create_port(x_width=self._input_width, y_width=self._output_width)

    def save_to(self, destination: Path) -> None:
        process_content = []

        pairs = self._compute_io_pairs()
        input_value, output_value = pairs[0]
        process_content.append(
            f"if signed_x <= to_signed({input_value}, BITWIDTH_OUTPUT) then "
            f"signed_y <= to_signed({output_value}, BITWIDTH_OUTPUT);"
        )
        for input_value, output_value in pairs[1:-1]:
            process_content.append(
                f"elsif signed_x <= to_signed({input_value}, BITWIDTH_OUTPUT) then "
                f"signed_y <= to_signed({output_value}, BITWIDTH_OUTPUT);"
            )
        _, output = pairs[-1]
        process_content.append(
            f"else signed_y <= to_signed({output}, BITWIDTH_OUTPUT);"
        )
        process_content.append("end if;")

        self._template.parameters.update(process_content=process_content)
        destination.create_subpath(self.name).as_file(".vhd").write(self._template)


def _assert_value_is_representable_with_n_bits(value: int, n_bits: int) -> None:
    min_value = 2 ** (n_bits - 1) * (-1)
    max_value = 2 ** (n_bits - 1) - 1

    if value < min_value or value > max_value:
        raise ValueError(
            f"The value '{value}' cannot be represented "
            f"with {n_bits} in two's complement representation."
        )



from elasticai.creator.file_generation.on_disk_path import OnDiskPath
from elasticai.creator.nn import Sequential
from elasticai.creator.nn.fixed_point import Linear, ReLU
from elasticai.creator.vhdl.system_integrations.firmware_env5 import FirmwareENv5

if __name__ == "__main__":
    inputs = 2
    outputs = 2
    total_bits = 8
    frac_bits = 4

    output_dir = "build_dir"
    destination = OnDiskPath(output_dir)

    model = Sequential(
        Linear(
            in_features=inputs,
            out_features=outputs,
            total_bits=total_bits,
            frac_bits=frac_bits,
        ),
        ReLU(total_bits=total_bits),
    )
    my_design = model.create_design("myNetwork")
    my_design.save_to(destination.create_subpath("srcs"))

    firmware = FirmwareENv5(
        network=my_design,
        x_num_values=inputs,
        y_num_values=outputs,
        id=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15],
        skeleton_version="v2",
    )
    firmware.save_to(destination)



"""Provides ways to use and extend the elasticai.creator plugin system.

The plugin systems evolves around the

* [`PluginLoader`](#elasticai.creator.plugin.PluginLoader) class
* [`PluginSymbol`](#elasticai.creator.plugin.PluginSymbol)protocol
* `meta.toml` files that describe the plugins

For convenience many functions convert the dicts loaded from the `meta.toml` file into `PluginSpec` objects.
The `meta.toml` file needs to define the value of each field of the `PluginSpec` class.
The function [`read_plugin_dicts_from_package`](#elasticai.creator.plugin.read_plugin_dicts_from_package) will read all plugins from the `plugins` key in the `meta.toml` file of a package.

**Example of a minimal `meta.toml` file**
```toml
[[plugins]]
name = "minimal_plugin"
target_platform = "elastic-node-v5"
target_runtime = "vhdl"
version = "0.1"
api_version = "0.1"
```

The few minimal fields that a plugin is required to define shall allow
plugin loaders to decide how to treat the plugin this could mean to

* ignore the plugin
* forward it to another software component
* check if the plugin is not compatible with the current setup

The following table lists these required fields:

:::{list-table}
* - Field name
  - Type
  - Description
* - **name**
  - `str`
  - The name of the plugin, used to identify the plugin
* - **target platform**
  - `str`
  - A string describing the target platform for the plugin, ie.
    the lowering pass it should be loaded into.
    Currently there is no strict definition of the semantics of this string.
* - **target runtime**
  - `str`
  - A string the runtime context for the plugin.
    Currently there is no strict definition of the semantics of this string.
* - **version**
  - `str`
  - A version string in the form `major.minor.[patch]`.
    Specifies the version of the plugin, ie. if you introduce a new feature or fix a bug, you should usually increase the minor version.
* - **api_version**
  - `str`
  - The version of the plugin API (plugin system) that this plugin was developed against.
    This is used to check if the plugin is compatible with the current system.
:::


:::{warning}
The set of required fields and their semantics is experimental and likely to change in the future.
:::

The `PluginLoader` will read that description from the `meta.toml` file
in a given package and use a user provided function to decide which
symbols to load from which module.
Assuming that each of these symbols implements the `PluginSymbol` protocol,
it will then call `load_into` on each of these symbols with a `plugin_receiver`.
The `plugin_receiver` is provided by the user as well.

Most other classes defined in this module are supposed to increase usability and expressiveness.

"""

import importlib.resources as _res
from abc import abstractmethod
from collections.abc import Callable, Iterable, Iterator
from dataclasses import dataclass
from functools import partial, update_wrapper
from importlib import import_module as _import_module
from inspect import signature as _signature
from typing import Generic, ParamSpec, Protocol, TypeAlias, TypeVar, Union

import tomlkit as toml


@dataclass
class PluginSpec:
    """The specification of a plugin.

    Typically built by reading a dictionary from a toml file and
    building the spec using <<build_plugin_spec, `build_plugin_spec()`>>.
    The dataclass is only used to provide convenient access to the fields,
    support type checking and improve code readability.

    You can achieve your goals just as well with the `PluginDict` dictionary.
    That is defined as an alias for `dict[str, str | tuple[str, ...]]`.
    """

    name: str
    target_platform: str
    target_runtime: str
    version: str
    api_version: str
    package: str


PluginDict: TypeAlias = dict[str, Union[str, tuple[str, ...], "PluginDict"]]

_PlRecT = TypeVar("_PlRecT", contravariant=True)
_T = TypeVar("_T")


PluginSpecT = TypeVar("PluginSpecT", bound=PluginSpec)

P = ParamSpec("P")

_ReturnT = TypeVar("_ReturnT", covariant=True)


class PluginLoader(Generic[_PlRecT]):
    """Get a set of plugins from a package, use `extract_fn` to resolve the symbols and load them into the `plugin_receiver`.

    .Args
    * `fetch`: A function that extracts ``PluginSymbol``s from plugin dictionaries. Use this to decide, based on the plugin specs, which symbols to load from which modules. Use the `SymbolFetcherBuilder` to easily compose new `fetch` functions.

    * `plugin_receiver`: The object that will receive the loaded symbols. _Receive_ in this context means that the loader will call `symbol.load_into(plugin_receiver)` for each of these symbols.
    That allows symbols defined in the plugin to run code in the context of the plugin receiver.
    E.g., it could register a lowering function in a `LoweringPass` if that lowering pass is given as the receiver.
    """

    def __init__(self, fetch: "SymbolFetcher", plugin_receiver: _PlRecT) -> None:
        self._extract_fn = fetch
        self._receiver = plugin_receiver

    def _get_plugin_dicts(self, package: str) -> Iterable[PluginDict]:
        yield from read_plugin_dicts_from_package(package)

    def load_from_package(self, package: str) -> None:
        """load all plugins defined by the `meta.toml` file found in the package."""
        plugin_dicts = self._get_plugin_dicts(package)
        symbols = self._extract_fn(plugin_dicts)
        for loadable in symbols:
            loadable.load_into(self._receiver)


class PluginSymbol(Protocol[_PlRecT]):
    """A symbol that the `PluginLoader` can load into a receiver object.

    The receiver can be any object.
    """

    @abstractmethod
    def load_into(self, /, receiver: _PlRecT) -> None: ...


class SymbolFetcher(Protocol[_PlRecT]):
    """Fetches ``PluginSymbol``s for the `PluginLoader`."""

    @abstractmethod
    def __call__(
        self, data: Iterable[PluginDict]
    ) -> Iterator[PluginSymbol[_PlRecT]]: ...


_SymbolFetcherBuilderT = TypeVar("_SymbolFetcherBuilderT", bound="SymbolFetcherBuilder")


class SymbolFetcherBuilder(Generic[PluginSpecT, _PlRecT]):
    """Build a `SymbolFetcher` from simpler functions.

    The `SymbolFetcherBuilder` composes simpler functions into a
    `SymbolFetcher`.

    .Args
    * `spec_type`: The type of the plugin spec that the `SymbolFetcher` will build from the plugin dictionaries.
    """

    def __init__(self, spec_type: type[PluginSpecT]) -> None:
        self._fns_over_iterables: list[
            Callable[[Iterable[PluginSpecT]], Iterator[PluginSymbol[_PlRecT]]]
        ] = []
        self._fns: list[Callable[[PluginSpecT], Iterator[PluginSymbol[_PlRecT]]]] = []
        self._spec_type: type[PluginSpecT] = spec_type

    def add_fn_over_iter(
        self: _SymbolFetcherBuilderT,
        fn: Callable[[Iterable[PluginSpecT]], Iterator[PluginSymbol[_PlRecT]]],
    ) -> _SymbolFetcherBuilderT:
        """Add a function that will be called for each plugin spec."""
        self._fns_over_iterables.append(fn)
        return self

    def add_fn(
        self: _SymbolFetcherBuilderT,
        fn: Callable[[PluginSpecT], Iterator[PluginSymbol[_PlRecT]]],
    ) -> _SymbolFetcherBuilderT:
        """Add a function that will be called once for all plugin specs."""
        self._fns.append(fn)
        return self

    def build(self) -> SymbolFetcher[_PlRecT]:
        # The next three locs will decouple `fetcher` from the builder's state.
        # Without them the function will keep looking up these values
        # from the builder's namespace instead of the function closure.
        # But the builders state might have changed in the meantime.
        spec_type = self._spec_type
        fns = self._fns
        fns_over_iterables = self._fns_over_iterables

        def fetcher(data: Iterable[PluginDict]) -> Iterator[PluginSymbol[_PlRecT]]:
            specs = map(partial(build_plugin_spec, spec_type=spec_type), data)
            for plugin in specs:
                for fn in fns:
                    yield from fn(plugin)

            for iter_fn in fns_over_iterables:
                yield from iter_fn(specs)

        return fetcher


class PluginSymbolFn(PluginSymbol[_PlRecT], Generic[_PlRecT, P, _ReturnT], Protocol):
    """A `PluginSymbol` that is also a function/callable."""

    @abstractmethod
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> _ReturnT: ...


def import_symbols(module: str, names: Iterable[str]) -> Iterator[PluginSymbol]:
    """import names from a module and yield the resulting objects."""
    m = _import_module(module)
    for name in set(names):
        yield getattr(m, name)


def make_plugin_symbol(
    load_into: Callable[[_PlRecT], None], fn: Callable[P, _T]
) -> PluginSymbolFn[_PlRecT, P, _T]:
    """Turn two functions into a loadable and callable plugin symbol.

    .Args
    * `load_into`: executed when `PluginLoader` calls `load_into` on the Symbol
    * `fn`: wrapped function. Executed when the created `PluginSymbolFn` is called.

    An important use case for plugin symbols is to provide callable functions,
    e.g., for lowering passes. `make_plugin_symbol` eases creating
    plugin symbols for these functions.

    For an example look at the implementation of the `type_handler`
    decorators in xref::api:ir2vhdl.adoc[ir2vhdl].
    ```
    """

    class _PS(PluginSymbolFn[_PlRecT, P, _T]):
        def load_into(self, receiver):
            load_into(receiver)

        def __call__(self, *args: P.args, **kwargs: P.kwargs) -> _T:
            return fn(*args, **kwargs)

    wrapped = _PS()

    update_wrapper(wrapped, fn)
    return wrapped


def build_plugin_spec(d: PluginDict, spec_type: type[PluginSpecT]) -> PluginSpecT:
    """inspect spec_type and build an instance of it from the dictionary `d`.

    Missing field raise an error while extra fields will be ignored.
    """
    args = d
    s = _signature(spec_type)
    expected_params = set(s.parameters.keys())
    actual_params = set(args.keys())
    if expected_params != actual_params:
        if actual_params.intersection(expected_params).issubset(expected_params):
            raise MissingFieldError(
                expected_params.difference(actual_params), spec_type
            )
    bound = {k: args[k] for k in s.parameters.keys()}
    return spec_type(**bound)  # type: ignore


def read_plugin_dicts_from_package(package: str) -> Iterable[PluginDict]:
    """read the meta.toml file from the package returning the list of plugin dictionaries."""
    t = _res.files(package).joinpath("meta.toml")
    parsed: list[PluginDict] = []
    if t.is_file():
        with t.open("rb") as f:
            content = toml.load(f).unwrap()
            parsed.extend(content["plugins"])
    for d in parsed:
        d.update(dict(package=package))
    return parsed


class MissingFieldError(Exception):
    def __init__(self, field_names: set[str], plugin_type: type[PluginSpecT]):
        super().__init__(
            f"missing required fields {field_names} for plugin spec '{plugin_type.__qualname__}'\n\tAre you sure you are loading the correct plugin?\n\tIs the meta.toml file correct?"
        )



from typing import cast

import pytest
import torch

from elasticai.creator.nn.float.round_to_float import RoundToFloat
from tests.tensor_test_case import assertTensorEqual


def roundToFloat(
    inputs: float | list[float] | torch.Tensor, mantissa_bits: int, exponent_bits: int
) -> torch.Tensor:
    if not isinstance(inputs, torch.Tensor):
        inputs = torch.tensor(inputs)
    return cast(torch.Tensor, RoundToFloat.apply(inputs, mantissa_bits, exponent_bits))


@pytest.mark.parametrize(
    "decimal,target_float,mantissa_bits,exponent_bits",
    [
        (0.0, 0.001953125, 6, 3),
        (9.5694284, 9.625, 6, 3),
        (-9.5694284, -9.625, 6, 3),
        (0.0, 0.125, 3, 1),
        (1.525, 1.5, 3, 1),
    ],
)
def test_round_single_number_to_correct_float(
    decimal: float, target_float: float, mantissa_bits: int, exponent_bits: int
) -> None:
    actual_float = roundToFloat(decimal, mantissa_bits, exponent_bits).item()
    assert actual_float == target_float


def test_uses_different_scale_for_each_value_in_tensor() -> None:
    values = [-14.9, -5.4, 0.05, 8.123456789]
    target_floats = [-14.875, -5.375, 0.0498046875, 8.125]
    actual_floats = roundToFloat(values, 6, 3)
    assertTensorEqual(target_floats, actual_floats)


@pytest.mark.parametrize("decimal", [42.424242, -25.12345678])
def test_raises_error_if_value_out_of_bounds(decimal: float) -> None:
    with pytest.raises(ValueError):
        _ = roundToFloat(decimal, 6, 3)



import re
from collections.abc import Iterable, Iterator

import pytest

import elasticai.creator.plugin as p


def test_importing_plugin_with_missing_fields_raises_meaningful_error():
    config_from_file = {
        "api_version": "0.1",
        "version": "0.2",
        "package": "mypackage",
        "target_platform": "env5",
        "target_runtime": "vhdl",
        "new_unknown_field": "value",
    }
    with pytest.raises(
        p.MissingFieldError,
        match=re.escape(
            "missing required fields {'name'} for plugin spec 'PluginSpec'\n\tAre you sure you are loading the correct plugin?\n\tIs the meta.toml file correct?"
        ),
    ):
        p.build_plugin_spec(config_from_file, p.PluginSpec)


class MinimalPluginLoader(p.PluginLoader):
    def __init__(self, extract_fn: p.SymbolFetcher):
        super().__init__(extract_fn, self)

    def _get_plugin_dicts(self, package: str) -> Iterable[p.PluginDict]:
        dummy_dicts: list[p.PluginDict] = []
        return dummy_dicts


class DummyLoadable(p.PluginSymbol[p.PluginLoader]):
    loaded = False

    @classmethod
    def load_into(cls, loader: p.PluginLoader) -> None:
        cls.loaded = True


def test_plugin_loader_loads_plugins():
    def extract_fn(data: Iterable[p.PluginDict]) -> Iterator[p.PluginSymbol]:
        yield from p.import_symbols("tests.unit_tests.plugin_test", {"DummyLoadable"})

    DummyLoadable.loaded = False
    loader = MinimalPluginLoader(extract_fn)
    loader.load_from_package("")
    assert DummyLoadable.loaded



import cocotb
from cocotb.clock import Clock
from cocotb.triggers import RisingEdge, Timer

from elasticai.creator.testing.cocotb_prepare import read_testdata


@cocotb.test()
async def precomputed_test(dut):
    data = read_testdata(dut._name)
    clock_period_ns = 10
    dut.enable.value = 1  # Has no impact
    dut.clock.value = 0  # has no impact
    dut.x.value = 0

    cocotb.start_soon(Clock(dut.clock, period=clock_period_ns, units="ns").start())
    await Timer(4 * clock_period_ns, units="ns")
    await RisingEdge(dut.clock)

    chck = list()
    for sig_in, ref_out in zip(data["in"], data["out"]):
        dut.x.value = sig_in
        for _ in range(2):
            await RisingEdge(dut.clock)

        chck.append(dut.y.value.signed_integer in [ref_out - 1, ref_out, ref_out + 1])
        if not chck[-1]:
            print(
                f"x={dut.x.value.signed_integer} -> y_pred={dut.y.value.signed_integer}, y_true={ref_out}"
            )

        for _ in range(2):
            await RisingEdge(dut.clock)

    accuracy = sum(chck) / len(chck)
    limit = 0.65 if "hardsigmoid" in dut._name else 0.98
    assert accuracy >= limit, f"Accuracy of {accuracy * 100:.2f}%"



from collections.abc import Iterable, Iterator, Mapping
from typing import Hashable, TypeVar

from .graph import Graph
from .name_generation import NameRegistry


class DanglingEdgeError(Exception):
    def __init__(self, a, b):
        super().__init__(f"produced dangling edge {a} -> {b}")


class RewriteResult:
    def __init__(
        self,
        *,
        new_graph: Graph[str],
        pattern_to_original: dict[str, str],
        replacement_to_new: dict[str, str],
    ) -> None:
        self.new_graph = new_graph
        self.pattern_to_original = pattern_to_original
        self.replacement_to_new = replacement_to_new


TG = TypeVar("TG", bound=Hashable)
TR = TypeVar("TR", bound=Hashable)
TI = TypeVar("TI", bound=Hashable)

T = TypeVar("T", bound=Hashable)
TP = TypeVar("TP", bound=Hashable)


def get_rewriteable_matches(
    original: Graph[T], matches: Iterable[dict[TP, T]], interface_nodes: Iterable[TP]
) -> Iterator[dict[TP, T]]:
    """Yield all matches that do produce dangling edges and do not overlap with previous matches.

    The matches returned by this function are considerd safe to be rewritten in a single rewriting step in any order,
    without having to run an additional matching step and
    without producing dangling edges.

    :param original: The original graph.
    :param matches: The matches to check.
    :param interface_nodes: All nodes in the pattern that are belong to the interface. These nodes are considered to be preserved during rewriting. Thus, the edges connected to these nodes are not considered dangling.

    """
    interface_nodes = set(interface_nodes)
    matched_nodes_wo_interfaces: set[T] = set()

    for match in matches:
        if not produces_dangling_edge(original, match, interface_nodes):
            new_matched_nodes = set(match.values())
            new_matched_interface_nodes = {
                match[interface_node] for interface_node in interface_nodes
            }
            new_matched_nodes_wo_interfaces = (
                new_matched_nodes - new_matched_interface_nodes
            )

            if new_matched_nodes.isdisjoint(matched_nodes_wo_interfaces):
                matched_nodes_wo_interfaces.update(new_matched_nodes_wo_interfaces)
                yield match


def produces_dangling_edge(
    graph: Graph[T], match: dict[TP, T], interface_nodes: Iterable[TP]
) -> bool:
    """Check if there are dangling edges attached to non-interface nodes."""

    def is_dangling_edge(src: T, dst: T) -> bool:
        return src not in match.values() or dst not in match.values()

    def non_interface_nodes() -> Iterator[T]:
        for pn, gn in match.items():
            if pn not in interface_nodes:
                yield gn

    for gn in non_interface_nodes():
        for gp in graph.predecessors[gn]:
            if is_dangling_edge(gp, gn):
                return True
        for gs in graph.successors[gn]:
            if is_dangling_edge(gn, gs):
                return True
    return False


def rewrite(
    *,
    replacement: Graph[str],
    original: Graph[str],
    match: Mapping[str, str],
    lhs: Mapping[str, str],
    rhs: Mapping[str, str],
) -> tuple[Graph[str], dict[str, str]]:
    """Return new rewritten graph and a mapping of replacements to new nodes.

    The terminology here is based on the double pushout approach to graph rewriting.
    The algorithm will find the first occurence of pattern in the graph using the `match` function.
    Then, it will create a new graph by replacing the pattern with the replacement.
    The structures specified by `interface` are excluded from this replacement.
    Instead the interface is used as the "glue" between the original graph and the replacement.
    The functions `lhs` and `rhs` are used to identify the `interface` nodes in the `pattern` and `replacement` respectively.

    :::{important}
    * The `rhs` function is required to be injective. This means that each interface node should be mapped to a unique replacement node.
    * The algorithm will stop on the first match. If you need to replace multiple matches you have to call the `rewrite` function multiple times.
    * The nodes from `replacement` will be automatically renamed to avoid conflicts with the nodes in the original graph, i.e., if node `a` exists already we will add a new node `a_1`.
    :::


    :::{note}
    - The nodes from the replacement graph that aren't part of the interface will be
        automatically renamed to avoid conflicts with nodes in the original graph.
    - Interface nodes serve as connection points between the original graph and replacement.

    :::

    :param replacement: The graph that will replace the matched pattern in the original graph.
    :param graph: The original graph where the pattern will be replaced.
    :param match: A dictionary mapping nodes from pattern to nodes in the original graph.
    :param lhs: Maps interface nodes to their corresponding nodes in the pattern (Left Hand Side).
    :param rhs: Maps interface nodes to their corresponding nodes in the replacement (Right Hand Side).
    :return: A tuple containing the new graph after the rewrite operation and a dictionary mapping replacement node names to new node names.
    :raises ValueError: If the `rhs` function is not injective (when different interface nodes map to the same replacement node).
    :raises DanglingEdgeError: if there is an edge between an unmatched node and a matched non-interface node.
    """
    rhs_inversed = {rhs[node]: node for node in rhs}
    replacement_nodes_in_interface = set(rhs_inversed.keys())
    if len(set(rhs.keys())) != len(replacement_nodes_in_interface):
        raise ValueError(
            "Ensure the `rhs` function is injective. The `rhs` function should map each interface node to a unique replacement node."
        )
    interface_nodes_in_pattern = set(lhs.values())
    interface_nodes_in_graph = set(match[node] for node in interface_nodes_in_pattern)

    new_graph: Graph = original.new()
    nodes_to_be_removed = set(match.values()) - set(interface_nodes_in_graph)
    nodes_to_keep = set(original.nodes) - nodes_to_be_removed
    name_registry = SingleNewNameGenerator(NameRegistry().prepopulate(nodes_to_keep))
    new_replacement_names: dict[str, str] = {}

    for node in set(replacement.nodes):
        if node in replacement_nodes_in_interface:
            new_replacement_names[node] = match[lhs[rhs_inversed[node]]]
        else:
            new_replacement_names[node] = name_registry.get_name(node)

    def get_replacement_node_in_new_graph(node: str) -> str:
        if node in replacement_nodes_in_interface:
            return match[lhs[rhs_inversed[node]]]
        return new_replacement_names[node]

    def get_replacement_edge_for_new_graph(src: str, dst: str) -> tuple[str, str]:
        src = get_replacement_node_in_new_graph(src)
        dst = get_replacement_node_in_new_graph(dst)
        return src, dst

    for node in nodes_to_keep:
        new_graph.add_node(node)
        if node not in interface_nodes_in_graph:
            for ns in original.successors[node]:
                if ns in nodes_to_be_removed:
                    raise DanglingEdgeError(node, ns)
            for np in original.predecessors[node]:
                if np in nodes_to_be_removed:
                    raise DanglingEdgeError(np, node)

    for node in replacement.nodes:
        if node not in replacement_nodes_in_interface:
            new_graph.add_node(get_replacement_node_in_new_graph(node))

    for src, dst in original.iter_edges():
        if src in nodes_to_keep and dst in nodes_to_keep:
            new_graph.add_edge(src, dst)

    for src, dst in replacement.iter_edges():
        new_graph.add_edge(*get_replacement_edge_for_new_graph(src, dst))

    return new_graph, new_replacement_names


class SingleNewNameGenerator:
    def __init__(self, registry: NameRegistry) -> None:
        self._registry = registry
        self._reversed: dict[str, str] = {}
        self.memory: dict[str, str] = {}

    def get_name(self, name: str) -> str:
        if name in self.memory:
            return self.memory[name]
        new_name = self._registry.get_unique_name(name)
        self.memory[name] = new_name
        self._reversed[new_name] = name
        return new_name



import pytest

from elasticai.creator.arithmetic.fxp_converter import FxpConverter, FxpParams


@pytest.mark.parametrize(
    "total_bits, frac_bits, number, check",
    [
        (2, 0, 1, '"01"'),
        (2, 1, -1, '"11"'),
        (4, 2, 1, '"0001"'),
        (4, 2, -1, '"1111"'),
        (4, 2, -3, '"1101"'),
        (8, 4, -98, '"10011110"'),
        (8, 4, 35, '"00100011"'),
    ],
)
def test_convert_integer_to_binary_vhdl(
    total_bits: int, frac_bits: int, number: int, check: str
):
    rslt = FxpConverter(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    ).integer_to_binary_string_vhdl(number)
    assert rslt == check


@pytest.mark.parametrize(
    "total_bits, frac_bits, number, check",
    [
        (2, 0, 1, 'X"1"'),
        (2, 1, -1, 'X"3"'),
        (4, 2, 1, 'X"1"'),
        (4, 2, -1, 'X"F"'),
        (4, 2, -3, 'X"D"'),
        (8, 4, -99, 'X"9D"'),
        (8, 4, 35, 'X"23"'),
    ],
)
def test_convert_integer_to_hex_vhdl(
    total_bits: int, frac_bits: int, number: int, check: str
):
    rslt = FxpConverter(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits)
    ).integer_to_hex_string_vhdl(number)
    assert rslt == check


@pytest.mark.parametrize(
    "total_bits, frac_bits, number, check",
    [
        (2, 0, 1.0, '"01"'),
        (2, 1, -1.0, '"10"'),
        (4, 2, 1.0, '"0100"'),
        (4, 2, -1.0, '"1100"'),
        (4, 2, -0.75, '"1101"'),
        (8, 4, -4.6, '"10110110"'),
        (8, 4, 2.4, '"00100110"'),
    ],
)
def test_convert_rational_to_binary_vhdl(
    total_bits: int, frac_bits: int, number: int, check: str
):
    rslt = FxpConverter(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits)
    ).rational_to_binary_string_vhdl(number)
    assert rslt == check


@pytest.mark.parametrize(
    "total_bits, frac_bits, number, check",
    [
        (2, 0, 1.0, 'X"1"'),
        (2, 1, -1.0, 'X"2"'),
        (4, 2, 1.0, 'X"4"'),
        (4, 2, -1.0, 'X"C"'),
        (4, 2, -0.75, 'X"D"'),
        (8, 4, -4.6, 'X"B6"'),
        (8, 4, 2.4, 'X"26"'),
    ],
)
def test_convert_rational_to_hex_vhdl(
    total_bits: int, frac_bits: int, number: int, check: str
):
    rslt = FxpConverter(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits)
    ).rational_to_hex_string_vhdl(number)
    assert rslt == check


@pytest.mark.parametrize(
    "total_bits, frac_bits, number, check",
    [
        (2, 0, 1, "2'b01"),
        (2, 1, -1, "2'b11"),
        (4, 2, 1, "4'b0001"),
        (4, 2, -1, "4'b1111"),
        (4, 2, -3, "4'b1101"),
        (8, 4, 103, "8'b01100111"),
        (8, 6, -7, "8'b11111001"),
    ],
)
def test_convert_integer_to_binary_verilog(
    total_bits: int, frac_bits: int, number: int, check: str
):
    rslt = FxpConverter(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits)
    ).integer_to_binary_string_verilog(number)
    assert rslt == check


@pytest.mark.parametrize(
    "total_bits, frac_bits, number, check",
    [
        (2, 0, 1, "2'd1"),
        (2, 1, -1, "2'd3"),
        (4, 2, 1, "4'd1"),
        (4, 2, -1, "4'd15"),
        (4, 2, -3, "4'd13"),
        (8, 4, -101, "8'd155"),
        (8, 4, 35, "8'd35"),
    ],
)
def test_convert_integer_to_decimal_verilog(
    total_bits: int, frac_bits: int, number: int, check: str
):
    rslt = FxpConverter(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits)
    ).integer_to_decimal_string_verilog(number)
    assert rslt == check


@pytest.mark.parametrize(
    "total_bits, frac_bits, number, check",
    [
        (2, 0, 1, "2'h1"),
        (2, 1, -1, "2'h3"),
        (4, 2, 1, "4'h1"),
        (4, 2, -1, "4'hF"),
        (4, 2, -3, "4'hD"),
        (8, 4, -101, "8'h9B"),
        (8, 4, 35, "8'h23"),
    ],
)
def test_convert_integer_to_hex_verilog(
    total_bits: int, frac_bits: int, number: int, check: str
):
    rslt = FxpConverter(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits)
    ).integer_to_hex_string_verilog(number)
    assert rslt == check


@pytest.mark.parametrize(
    "total_bits, frac_bits, number, check",
    [
        (2, 0, 1, "2'b01"),
        (2, 1, -0.5, "2'b11"),
        (4, 2, 0.25, "4'b0001"),
        (4, 2, -0.25, "4'b1111"),
        (4, 2, -0.75, "4'b1101"),
        (8, 4, 1.75, "8'b00011100"),
        (8, 6, -0.4345, "8'b11100100"),
    ],
)
def test_convert_rational_to_binary_verilog(
    total_bits: int, frac_bits: int, number: int, check: str
):
    rslt = FxpConverter(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits)
    ).rational_to_binary_string_verilog(number)
    assert rslt == check


@pytest.mark.parametrize(
    "total_bits, frac_bits, number, check",
    [
        (2, 0, 1, "2'd1"),
        (2, 1, -0.5, "2'd3"),
        (4, 2, 0.25, "4'd1"),
        (4, 2, -0.25, "4'd15"),
        (4, 2, -0.75, "4'd13"),
        (8, 4, 1.75, "8'd28"),
        (8, 6, -0.4345, "8'd228"),
    ],
)
def test_convert_rational_to_decimal_verilog(
    total_bits: int, frac_bits: int, number: float, check: str
):
    rslt = FxpConverter(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits)
    ).rational_to_decimal_string_verilog(number)
    assert rslt == check


@pytest.mark.parametrize(
    "total_bits, frac_bits, number, check",
    [
        (2, 0, 1, "2'h1"),
        (2, 1, -0.5, "2'h3"),
        (4, 2, 0.25, "4'h1"),
        (4, 2, -0.25, "4'hF"),
        (4, 2, -0.75, "4'hD"),
        (8, 4, 1.75, "8'h1C"),
        (8, 6, -0.4345, "8'hE4"),
    ],
)
def test_convert_rational_to_hex_verilog(
    total_bits: int, frac_bits: int, number: float, check: str
):
    rslt = FxpConverter(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits)
    ).rational_to_hex_string_verilog(number)
    assert rslt == check


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, number, expected",
    [
        (2, 0, True, "01", 1),
        (2, 0, False, "11", 3),
        (2, 1, True, "10", -2),
        (4, 2, True, "0100", 4),
        (4, 2, True, "1100", -4),
        (4, 0, True, "1100", -4),
        (4, 0, False, "1100", 12),
        (4, 2, True, "1101", -3),
        (8, 4, True, "10110110", -74),
        (8, 4, True, "00100110", 38),
    ],
)
def test_convert_binary_vhdl_to_integer(
    total_bits: int, frac_bits: int, signed: bool, number: str, expected: str
):
    rslt = FxpConverter(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=signed)
    ).binary_to_integer(number)
    assert rslt == expected


@pytest.mark.parametrize(
    "total_bits, frac_bits, signed, number, expected",
    [
        (2, 0, True, '"01"', 1.0),
        (2, 0, False, '"11"', 3.0),
        (2, 1, True, '"10"', -1.0),
        (4, 2, True, '"0100"', 1.0),
        (4, 2, True, "4'b0100", 1.0),
        (4, 2, True, '"1100"', -1.0),
        (4, 2, True, '"1101"', -0.75),
        (4, 2, True, "4'b1101", -0.75),
        (8, 4, True, '"10110110"', -4.625),
        (8, 4, True, '"00100110"', 2.375),
        (8, 4, True, "8'b00100110", 2.375),
    ],
)
def test_convert_binary_vhdl_to_float(
    total_bits: int, frac_bits: int, signed: bool, number: str, expected: str
):
    rslt = FxpConverter(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=signed)
    ).binary_to_rational(number)
    assert rslt == expected



from elasticai.creator.function_utils import FunctionDecorator


def test_can_call_with_name_and_fn():
    outer_name = ""

    def callback(name: str, fn):
        nonlocal outer_name
        outer_name = name
        return fn

    decorate = FunctionDecorator(callback)

    def fn():
        return "x"

    fn = decorate("fn_0", fn)
    assert "x" == fn()
    assert outer_name == "fn_0"



import torch

from elasticai.creator.nn.float.math_operations import MathOperations
from tests.tensor_test_case import assertTensorEqual


def test_quantize() -> None:
    arithmetics = MathOperations(mantissa_bits=3, exponent_bits=1)
    a = torch.tensor([-3.0, -2.0, -1.69, -0.2, 0.2, 1.69, 2.0, 3.0])
    assertTensorEqual(
        expected=[-1.875, -1.875, -1.75, -0.203125, 0.203125, 1.75, 1.875, 1.875],
        actual=arithmetics.quantize(a),
    )


def test_add() -> None:
    arithmetics = MathOperations(mantissa_bits=3, exponent_bits=1)
    a = torch.tensor([-0.6875, 0.1250, 0.6875])
    b = torch.tensor([-0.3125, 0.2031, 0.3125])
    assertTensorEqual(
        expected=[-1.0, 0.3125, 1.0],
        actual=arithmetics.add(a, b),
    )


def test_mul() -> None:
    arithmetics = MathOperations(mantissa_bits=3, exponent_bits=1)
    a = torch.tensor([-0.6875, 0.1250, 0.6875])
    b = torch.tensor([-0.3125, 0.2031, 0.3125])
    assertTensorEqual(
        expected=[0.21875, 0.125, 0.21875],
        actual=arithmetics.mul(a, b),
    )


def test_matmul() -> None:
    arithmetics = MathOperations(mantissa_bits=3, exponent_bits=1)
    a = torch.tensor([[-2.0, -1.75, -1.5], [-0.25, 0.0, 0.25], [1.25, 1.5, 1.75]])
    b = torch.tensor([[-0.25], [0.5], [0.25]])
    assertTensorEqual(
        expected=[[-0.75], [0.125], [0.875]],
        actual=arithmetics.matmul(a, b),
    )



import pytest

from elasticai.creator.file_generation.in_memory_path import InMemoryFile, InMemoryPath
from elasticai.creator.nn.fixed_point.conv1d import Conv1d
from elasticai.creator.nn.sequential import Sequential
from elasticai.creator.vhdl.design.design import Design


@pytest.fixture
def seq_conv1d_design() -> Design:
    return Sequential(
        Conv1d(
            total_bits=16,
            frac_bits=8,
            in_channels=1,
            out_channels=2,
            kernel_size=3,
            signal_length=4,
        )
    ).create_design("network")


def save_design(design: Design) -> dict[str, str]:
    destination = InMemoryPath("network", parent=None)
    design.save_to(destination)
    files = recursive_children_extraction(destination.children.values())
    return files


def recursive_children_extraction(
    destination: dict[str, InMemoryPath | InMemoryFile],
) -> dict[str, str | dict]:
    data = {}
    for dest in destination:
        if isinstance(dest, InMemoryFile):
            data[dest.name] = "\n".join(dest.text)
        elif type(dest == InMemoryPath):
            data[dest.name] = recursive_children_extraction(dest.children.values())
    return data


def test_network_generation_1(seq_conv1d_design: Design):
    saved_files = save_design(seq_conv1d_design)
    print(saved_files)
    expected_files = ["conv1d_0", "network.vhd"]
    actual_files = list(saved_files.keys())
    assert expected_files == actual_files


def test_network_generation_2(seq_conv1d_design: Design):
    saved_files = save_design(seq_conv1d_design)
    expected_files_in_conv1d_0 = [
        "conv1d_0_w_rom.vhd",
        "conv1d_0_b_rom.vhd",
        "conv1d_0.vhd",
        "conv1d_0_fxp_MAC_RoundToZero.vhd",
        "fxp_mac.vhd",
    ]
    actual_files_in_conv1d0 = list(saved_files["conv1d_0"].keys())
    print(expected_files_in_conv1d_0)
    print(actual_files_in_conv1d0)
    for item in actual_files_in_conv1d0:
        assert item in expected_files_in_conv1d_0



import torch
from torch import Tensor
from torch.nn import Module

from .autograd import (
    QuantizeForwHTEAutograd,
    QuantizeForwHTEBackwHTEAutograd,
    QuantizeForwHTEBackwStochasticAutograd,
    QuantizeForwStochasticAutograd,
    QuantizeForwStochasticBackwStochasticAutograd,
)
from .two_complement_fixed_point_config import FixedPointConfigV2


class _ForwModule(Module):
    def __init__(self, forward_conf: FixedPointConfigV2): ...


class _ForwBackwModule(Module):
    def __init__(
        self, forward_conf: FixedPointConfigV2, backward_conf: FixedPointConfigV2
    ): ...


def get_fxp_forw_quantization_module(
    autograd_func: torch.autograd.Function,
) -> type[_ForwModule]:
    class FxPForwQuantizationModule(Module):
        def __init__(self, forward_conf: FixedPointConfigV2) -> None:
            super().__init__()
            self.autograd_func = autograd_func
            self.register_buffer(
                "forw_resolution_per_int", forward_conf.resolution_per_int
            )
            self.register_buffer(
                "forw_minimum_as_rational", forward_conf.minimum_as_rational
            )
            self.register_buffer(
                "forw_maximum_as_rational", forward_conf.maximum_as_rational
            )

        def forward(self, x: Tensor) -> Tensor:
            return self.autograd_func.apply(
                x,
                self.forw_resolution_per_int,
                self.forw_minimum_as_rational,
                self.forw_maximum_as_rational,
            )

    return FxPForwQuantizationModule


def get_fxp_forwbackw_quantization_module(
    autograd_func: torch.autograd.Function,
) -> type[_ForwBackwModule]:
    class FxPForwBackwQuantizationModule(Module):
        def __init__(
            self, forward_conf: FixedPointConfigV2, backward_conf: FixedPointConfigV2
        ) -> None:
            super().__init__()
            self.autograd_func = autograd_func
            self.register_buffer(
                "forw_resolution_per_int", forward_conf.resolution_per_int
            )
            self.register_buffer(
                "forw_minimum_as_rational", forward_conf.minimum_as_rational
            )
            self.register_buffer(
                "forw_maximum_as_rational", forward_conf.maximum_as_rational
            )
            self.register_buffer(
                "backw_resolution_per_int", backward_conf.resolution_per_int
            )
            self.register_buffer(
                "backw_minimum_as_rational", backward_conf.minimum_as_rational
            )
            self.register_buffer(
                "backw_maximum_as_rational", backward_conf.maximum_as_rational
            )
            self._kwargs = {
                "forw_resolution_per_int": self.forw_resolution_per_int,
                "forw_minimum_as_rational": self.forw_minimum_as_rational,
                "forw_maximum_as_rational": self.forw_maximum_as_rational,
                "backw_resolution_per_int": self.backw_resolution_per_int,
                "backw_minimum_as_rational": self.backw_minimum_as_rational,
                "backw_maximum_as_rational": self.backw_maximum_as_rational,
            }

        def forward(self, x: Tensor) -> Tensor:
            return self.autograd_func.apply(
                x,
                self.forw_resolution_per_int,
                self.forw_minimum_as_rational,
                self.forw_maximum_as_rational,
                self.backw_resolution_per_int,
                self.backw_minimum_as_rational,
                self.backw_maximum_as_rational,
            )

    return FxPForwBackwQuantizationModule


QuantizeForwHTE: type[_ForwModule] = get_fxp_forw_quantization_module(
    QuantizeForwHTEAutograd
)
QuantizeForwStochastic: type[_ForwModule] = get_fxp_forw_quantization_module(
    QuantizeForwStochasticAutograd
)
QuantizeForwHTEBackwHTE: type[_ForwBackwModule] = get_fxp_forwbackw_quantization_module(
    QuantizeForwHTEBackwHTEAutograd
)
QuantizeForwHTEBackwStochastic: type[_ForwBackwModule] = (
    get_fxp_forwbackw_quantization_module(QuantizeForwHTEBackwStochasticAutograd)
)
QuantizeForwStochasticBackwStochastic: type[_ForwBackwModule] = (
    get_fxp_forwbackw_quantization_module(QuantizeForwStochasticBackwStochasticAutograd)
)



from typing import Iterable, Iterator, Sequence, TypeVar, overload

T = TypeVar("T")


@overload
def batched(vector: Sequence[T], size: int) -> Iterator[Sequence[T]]: ...


@overload
def batched(vector: Iterable[T], size: int) -> Iterator[Iterable[T]]: ...


def batched(vector, size):
    def make_batches(vector, size):
        steps = len(vector) // size
        for step in range(steps):
            yield vector[size * step : size * (step + 1)]

    def make_sequence(vector):
        is_iterable = hasattr(vector, "__iter__")
        is_sequence = hasattr(vector, "__len__") and hasattr(vector, "__getitem__")
        if not is_iterable and not is_sequence:
            raise ValueError("argument needs to be a Sequence or Iterable")
        if not is_sequence:
            return tuple(vector)
        return vector

    def handle_invalid_length(vector, size):
        if len(vector) % size != 0:
            raise ValueError

    vector = make_sequence(vector)
    handle_invalid_length(vector, size)
    yield from make_batches(vector, size)



from dataclasses import dataclass
from typing import Protocol, TypeVar, Union, cast, overload, runtime_checkable

T = TypeVar("T", bound="ConvertableToFixedPointValues")


@runtime_checkable
class ConvertableToFixedPointValues(Protocol[T]):
    def round(self: T) -> T: ...

    def int(self: T) -> T: ...

    def float(self: T) -> T: ...

    def __gt__(self: T, other: Union[int, float, T]) -> T:  # type: ignore
        ...

    def __lt__(self: T, other: Union[int, float, T]) -> T:  # type: ignore
        ...

    def __or__(self: T, other: T) -> T: ...

    def __mul__(self: T, other: Union[int, T, float]) -> T:  # type: ignore
        ...

    def __truediv__(self: T, other: Union[int, float]) -> T:  # type: ignore
        ...


@dataclass(frozen=True)
class FxpParams:
    total_bits: int
    frac_bits: int
    signed: bool = True

    def __post_init__(self):
        if self.total_bits <= 0:
            raise Exception(
                f"total bits need to be > 0 for {self.__class__.__name__}. "
                f"You have set {self.total_bits=}."
            )
        if self.frac_bits > self.total_bits:
            raise Exception(
                f"total bits-1 needs to be > frac bits for {self.__class__.__name__}. "
                f"You have set {self.total_bits=} and {self.frac_bits=}."
            )

    @property
    def minimum_as_integer(self) -> int:
        return 2 ** (self.total_bits - 1) * (-1) if self.signed else 0

    @property
    def maximum_as_integer(self) -> int:
        return 2 ** (self.total_bits - 1) - 1 if self.signed else 2**self.total_bits - 1

    @property
    def minimum_as_rational(self) -> float:
        return self.minimum_as_integer * self.minimum_step_as_rational

    @property
    def minimum_step_as_rational(self) -> float:
        return 1 / (1 << self.frac_bits)

    @property
    def maximum_as_rational(self) -> float:
        return self.maximum_as_integer * self.minimum_step_as_rational

    @overload
    def integer_out_overflow(self, number: T) -> T: ...

    @overload
    def integer_out_overflow(self, number: int) -> bool: ...

    def integer_out_overflow(self, number: int | T) -> bool | T:
        return number > self.maximum_as_integer

    @overload
    def integer_out_underflow(self, number: T) -> T: ...

    @overload
    def integer_out_underflow(self, number: int) -> bool: ...

    def integer_out_underflow(self, number: int | T) -> bool | T:
        return number < self.minimum_as_integer

    @overload
    def integer_out_of_bounds(self, number: T) -> T: ...

    @overload
    def integer_out_of_bounds(self, number: int) -> bool: ...

    def integer_out_of_bounds(self, number: int | T) -> bool | T:
        if isinstance(number, ConvertableToFixedPointValues):
            return self._check_integer_out_of_bounds(cast(T, number))
        else:
            return self._check_integer_out_of_bounds(number)

    def _check_integer_out_of_bounds(self, number) -> bool | T:
        return self.integer_out_underflow(number) | self.integer_out_overflow(number)

    @overload
    def rational_out_overflow(self, number: T) -> T: ...

    @overload
    def rational_out_overflow(self, number: float) -> bool: ...

    def rational_out_overflow(self, number: float | T) -> bool | T:
        return number > self.maximum_as_rational

    @overload
    def rational_out_underflow(self, number: T) -> T: ...

    @overload
    def rational_out_underflow(self, number: float) -> bool: ...

    def rational_out_underflow(self, number: float | T) -> bool | T:
        return number < self.minimum_as_rational

    @overload
    def rational_out_of_bounds(self, number: T) -> T: ...

    @overload
    def rational_out_of_bounds(self, number: float) -> bool: ...

    def rational_out_of_bounds(self, number) -> bool | T:
        return self.rational_out_underflow(number) | self.rational_out_overflow(number)



import pytest

import elasticai.creator.graph as gr
import elasticai.creator.ir as ir
from elasticai.creator.ir.rewriting.rewriter import RemappedSubImplementation


class Node(ir.Node):
    size: int


class TestRemapping:
    @pytest.fixture
    def replaced(self) -> ir.Implementation:
        return (
            ir.Implementation(graph=gr.BaseGraph(), data={})
            .add_nodes(
                (
                    ir.node(name="a", type="ta"),
                    ir.node(name="b", type="tb"),
                    ir.node(name="c", type="tc"),
                )
            )
            .add_edges((ir.edge("a", "b"), ir.edge("b", "c"), ir.edge("a", "c")))
        )

    @pytest.fixture
    def replacement(self):
        return (
            ir.Implementation(graph=gr.BaseGraph(), data={})
            .add_nodes(
                (
                    ir.node(name="x", type="ta"),
                    ir.node(name="y", type="tb"),
                    ir.node(name="z", type="tc"),
                )
            )
            .add_edges((ir.edge("x", "y"), ir.edge("y", "z"), ir.edge("x", "z")))
        )

    @pytest.fixture
    def remapped(
        self,
        replaced: ir.Implementation[ir.Node, ir.Edge],
        replacement: ir.Implementation[ir.Node, ir.Edge],
    ):
        return RemappedSubImplementation(
            mapping={"x": "a", "y": "b", "z": "c"},
            graph=replacement.graph,
            data=replaced.data["nodes"],  #  type: ignore
            node_fn=Node,
            edge_fn=lambda src, dst, data: ir.Edge(src, dst, data),
        )

    def test_change_node_data_using_names_from_replacement(
        self,
        remapped: RemappedSubImplementation[Node, ir.Edge],
        replaced: ir.Implementation,
    ):
        replacement_node_name = "x"
        corresponding_node_name_in_resulting_graph = "a"
        remapped.nodes[replacement_node_name].size = 4
        assert (
            replaced.nodes[corresponding_node_name_in_resulting_graph].attributes[
                "size"
            ]
            == 4
        )



from typing import cast

import pytest

from elasticai.creator.file_generation.in_memory_path import InMemoryFile, InMemoryPath
from elasticai.creator.nn.fixed_point.linear.design import LinearDesign


@pytest.fixture
def linear_design() -> LinearDesign:
    return LinearDesign(
        name="linear",
        in_feature_num=3,
        out_feature_num=2,
        total_bits=16,
        frac_bits=8,
        weights=[[1, 3, 5], [2, 4, 6]],
        bias=[7, 8],
    )


def save_design(design: LinearDesign) -> dict[str, str]:
    destination = InMemoryPath("linear", parent=None)
    design.save_to(destination)
    files = cast(list[InMemoryFile], list(destination.children.values()))
    return {file.name: "\n".join(file.text) for file in files}


def test_saved_design_contains_needed_files(linear_design: LinearDesign) -> None:
    saved_files = save_design(linear_design)
    expected_files = {"linear_rom.vhd", "linear.vhd"}
    actual_files = set(saved_files.keys())
    assert expected_files == actual_files


def test_linear_rom_code_generated_correctly(linear_design: LinearDesign) -> None:
    expected_code = """library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;

entity linear_rom is
    generic (
        ROM_ADDR_WIDTH : integer := 3;
        ROM_DATA_WIDTH : integer := 16
    );
    port (
        clk : in std_logic;
        en : in std_logic;
        addr : in std_logic_vector(ROM_ADDR_WIDTH-1 downto 0);
        data : out std_logic_vector(ROM_DATA_WIDTH-1 downto 0)
    );
end entity linear_rom;
architecture rtl of linear_rom is
    type linear_rom_array_t is array (0 to 2**ROM_ADDR_WIDTH-1) of std_logic_vector(ROM_DATA_WIDTH-1 downto 0);
    signal ROM : linear_rom_array_t:=("0000000000000001", "0000000000000011", "0000000000000101", "0000000000000111", "0000000000000010", "0000000000000100", "0000000000000110", "0000000000001000");
    attribute rom_style : string;
    attribute rom_style of ROM : signal is "auto";
begin
    ROM_process: process(addr)
    begin
        if (en = '1') then
            data <= ROM(to_integer(unsigned(addr)));
        end if;
    end process ROM_process;
end architecture rtl;"""
    saved_files = save_design(linear_design)
    actual_code = saved_files["linear_rom.vhd"]
    print(actual_code)
    assert expected_code == actual_code



from functools import partial

from elasticai.creator.nn.fixed_point.linear.layer import Linear as FPLinear1d

from ._common_imports import (
    Design,
    InProjectTemplate,
    Path,
    Port,
    Signal,
    calculate_address_width,
    module_to_package,
    std_signals,
)


class LSTMNetworkDesign(Design):
    def __init__(
        self,
        lstm: Design,
        linear_layer: FPLinear1d,
        total_bits: int,
        frac_bits: int,
        hidden_size: int,
        input_size: int,
    ) -> None:
        super().__init__(name="lstm_network")
        self._linear_layer = linear_layer
        self._lstm = lstm
        self.template = InProjectTemplate(
            module_to_package(self.__module__),
            file_name="lstm_network.tpl.vhd",
            parameters=dict(
                name=self.name,
                lstm_cell_name=self._lstm.name,
                linear_name=self._linear_layer.name,
                data_width=str(total_bits),
                frac_width=str(frac_bits),
                hidden_size=str(hidden_size),
                input_size=str(input_size),
                linear_in_features=str(self._linear_layer.in_feature_num),
                linear_out_features=str(self._linear_layer.out_feature_num),
                hidden_addr_width=(
                    f"{calculate_address_width(hidden_size + input_size)}"
                ),
                x_h_addr_width=f"{calculate_address_width(hidden_size + input_size)}",
                w_addr_width=(
                    f"{calculate_address_width((hidden_size + input_size) * hidden_size)}"
                ),
                in_addr_width="4",
            ),
        )

        ctrl_signal = partial(Signal, width=0)
        self._port = Port(
            incoming=[
                std_signals.clock(),
                std_signals.enable(),
                ctrl_signal("x_we"),
                Signal("x", width=lstm.port["x_data"].width),
                Signal("addr_in", width=lstm.port["h_out_addr"].width),
            ],
            outgoing=[
                std_signals.done(),
                Signal("d_out", width=lstm.port["h_out_data"].width),
            ],
        )
        self._subpath_name = "lstm_network"

    @property
    def port(self) -> Port:
        return self._port

    def save_to(self, destination: Path) -> None:
        self._lstm.save_to(destination)
        self._linear_layer.save_to(destination.create_subpath(self._linear_layer.name))
        destination.create_subpath(self._subpath_name).as_file(".vhd").write(
            self.template
        )



__all__ = [
    "IrData",
    "ReadOnlyField",
    "ReadOnlyMethodField",
    "StaticMethodField",
    "read_only_field",
    "static_required_field",
    "RequiredField",
    "SimpleRequiredField",
    "Attribute",
]

from .attribute import Attribute
from .ir_data import IrData
from .required_field import (
    ReadOnlyField,
    ReadOnlyMethodField,
    RequiredField,
    SimpleRequiredField,
    StaticMethodField,
    read_only_field,
    static_required_field,
)



from .fxp_arithmetic import FxpArithmetic as FxpArithmetic
from .fxp_converter import FxpConverter as FxpConverter
from .fxp_params import FxpParams as FxpParams

__all__ = ["FxpParams", "FxpArithmetic", "FxpConverter"]



from collections.abc import Iterable

from elasticai.creator.ir2vhdl import Shape
from elasticai.creator.ir2vhdl import VhdlNode as Node
from elasticai.creator_plugins.combinatorial.vhdl_nodes import shift_register


def extract_code_section(lines: Iterable[str], start: str, end: str):
    section_is_relevant = False
    lines = map(str.strip, lines)
    lines = tuple(lines)
    for line in lines:
        if not section_is_relevant and line.startswith(start):
            section_is_relevant = True
        elif section_is_relevant and line.startswith(end):
            section_is_relevant = False
        elif section_is_relevant:
            yield line


def test_shift_register_converts_depth_and_width_to_correct_generics():
    conv0_channels = 2
    conv1_kernel_size = 3
    conv0_out_shape = Shape(conv0_channels, 1)
    conv1_in_shape = Shape(conv0_channels, conv1_kernel_size)
    n = Node(
        name="sr0",
        data=dict(
            type="shift_register",
            implementation="",
            input_shape=conv0_out_shape.to_tuple(),
            output_shape=conv1_in_shape.to_tuple(),
        ),
    )
    sr = shift_register(n)
    entity = tuple(extract_code_section(sr.instantiate(), start="generic", end=")"))
    entity = set(line.strip(",") for line in entity)
    expected = {
        f"DATA_WIDTH => {conv0_channels}",
        f"NUM_POINTS => {conv1_in_shape.width}",
    }
    assert entity == expected



from importlib.metadata import version as _pypackage_version
from pathlib import Path

import tomlkit as toml
from tomlkit.toml_file import TOMLFile


class MetaFile:
    version = "0.1"

    def __init__(self, data):
        self._data = data

    @classmethod
    def load(cls, file: Path) -> "MetaFile":
        with open(file, "r") as f:
            data = toml.load(f)
        return cls(data)

    def set_skeleton_id(self, skeleton_id: bytes):
        self._data["skeleton_id"] = skeleton_id.hex()

    def save(self, file: Path) -> None:
        f = TOMLFile(file)
        f.write(self._data)

    @classmethod
    def default(cls) -> "MetaFile":
        creator_version = _pypackage_version("elasticai.creator")
        data = dict(
            meta_version=cls.version,
            creator_version=creator_version,
            hw_accelerator_version="0.1",
            skeleton_id='"<invalid>"  # this should be computed and set automatically',
            name="<your_network>",
            description="""An optional possibly very long description of what
                            your accelerator does and how to use it.""",
        )
        return cls(data)



import torch

from elasticai.creator_plugins.quantized_grads.fixed_point import FixedPointConfigV2
from elasticai.creator_plugins.quantized_grads.fixed_point.quantize_to_fixed_point import (
    _clamp,
    _round_to_fixed_point_hte,
    quantize_to_fxp_hte,
)


def test_round_determinstic_to_fixed_point():
    conf = FixedPointConfigV2(total_bits=8, frac_bits=2)
    x = torch.arange(-2, 2.1, step=0.1, dtype=torch.float32)

    actual = _round_to_fixed_point_hte(x, conf.resolution_per_int)
    expected = torch.Tensor(
        [
            -2.0,
            -2.0,
            -1.75,
            -1.75,
            -1.5,
            -1.5,
            -1.5,
            -1.25,
            -1.25,
            -1.0,
            -1.0,
            -1.0,
            -0.75,
            -0.75,
            -0.5,
            -0.5,
            -0.5,
            -0.25,
            -0.25,
            -0.0,
            0.0,
            0.0,
            0.25,
            0.25,
            0.5,
            0.5,
            0.5,
            0.75,
            0.75,
            1.0,
            1.0,
            1.0,
            1.25,
            1.25,
            1.5,
            1.5,
            1.5,
            1.75,
            1.75,
            2.0,
            2.0,
        ]
    )
    assert torch.equal(actual, expected)


def test_clamp_to_fixed_point():
    conf = FixedPointConfigV2(total_bits=4, frac_bits=2)
    x = torch.arange(-3, 4, dtype=torch.float32)

    actual = _clamp(
        number=x,
        minimum_as_rational=conf.minimum_as_rational,
        maximum_as_rational=conf.maximum_as_rational,
    )
    expected = torch.Tensor([-2.0, -2.0, -1.0, 0.0, 1.0, 1.75, 1.75])
    assert torch.equal(actual, expected)


def test_quantize_determinstic_to_fixed_point():
    conf = FixedPointConfigV2(total_bits=4, frac_bits=2)
    x = torch.arange(-3, 3.1, step=0.33, dtype=torch.float32)
    actual = quantize_to_fxp_hte(
        number=x,
        resolution_per_int=conf.resolution_per_int,
        minimum_as_rational=conf.minimum_as_rational,
        maximum_as_rational=conf.maximum_as_rational,
    )
    expected = torch.Tensor(
        [
            -2.0,
            -2.0,
            -2.0,
            -2.0,
            -1.75,
            -1.25,
            -1.0,
            -0.75,
            -0.25,
            -0.0,
            0.25,
            0.75,
            1.0,
            1.25,
            1.5,
            1.75,
            1.75,
            1.75,
            1.75,
        ]
    )
    assert torch.equal(actual, expected)



import json

import pytest

from elasticai.creator.graph import BaseGraph as _Graph
from elasticai.creator.ir import Attribute, Edge, Implementation, Node, edge, node


class Graph(Implementation[Node, Edge]):
    def __init__(self, data: dict[str, Attribute] | None = None):
        g: _Graph[str] = _Graph()
        if data is None:
            data = {}
        super().__init__(
            graph=g,
            node_fn=Node,
            edge_fn=Edge,
            data=data,
        )


def test_graph_is_serialized():
    g = Graph(data=dict(name="network", type="network"))
    g.add_node(node(name="node1", type="type1"))
    g.add_node(node(name="node2", type="type2"))
    g.add_edge(edge(src="node1", dst="node2"))
    assert g.data == {
        "name": "network",
        "type": "network",
        "nodes": {
            "node1": {"type": "type1"},
            "node2": {"type": "type2"},
        },
        "edges": {"node1": {"node2": {}}},
    }


def test_graph_has_required_fields():
    g = Graph(data=dict(name="network", type="network"))
    assert g.name == "network"


def test_can_retrieve_edges() -> None:
    g = (
        Graph()
        .add_node(name="x", data=dict(type="t"))
        .add_node(name="y", data=dict(type="t"))
        .add_edge(src="x", dst="y", data=dict())
    )
    edges = tuple(g.edges.values())
    assert edge(src="x", dst="y") == edges[0]


@pytest.fixture
def graph() -> Graph:
    g = (
        Graph()
        .add_node(name="x", data=dict(type="t"))
        .add_node(name="y", data=dict(type="t"))
        .add_node(name="z", data=dict(type="t"))
        .add_edge(src="x", dst="y", data={})
        .add_edge(src="y", dst="z", data=dict(extra="e"))
    )
    return g


def test_can_get_specific_edge(graph: Graph) -> None:
    e = graph.edges[("y", "z")]
    assert dict(extra="e") == e.attributes


def test_edges_is_read_only(graph: Graph) -> None:
    with pytest.raises(TypeError):
        graph.edges[("x", "z")] = edge("x", "z")  # type: ignore


def test_predecessors_of_z_is_y(graph: Graph) -> None:
    n = next(iter(graph.predecessors("z").values()))
    assert graph.nodes["y"] == n


def test_successor_of_x_is_y(graph: Graph) -> None:
    n = next(iter(graph.successors("x").values()))
    assert graph.nodes["y"] == n


def test_is_json_serializable(graph: Graph) -> None:
    result = json.dumps(graph.as_dict())
    assert result == (
        r'{"nodes": {"x": {"type": "t"}, "y": {"type": "t"}, "z": {"type": "t"}}, "edges": {"x": {"y": {}}, "y": {"z": {"extra": "e"}}}}'
    )


def test_change_node_attribute() -> None:
    impl = Implementation(graph=_Graph())
    impl.add_node(node(name="a", type="t"))
    impl.nodes["a"].data.update({"x": "y"})
    assert "y" == impl.nodes["a"].attributes["x"]
